## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.rl.envs
## -- Module  : doublependulum.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2022-01-25  0.0.0     WB       Creation
## -- 2022-01-26  0.9.0     WB       Initial trial of the environment
## -- 2022-01-27  0.9.1     WB       Trial without animation 
## -- 2022-01-28  0.9.2     WB       Fix the  update_plot method
## -- 2022-01-31  0.9.3     WB       Taking account of the new state in _compute_reward method 
## -- 2022-01-31  0.9.4     WB       Add Circular arrow to the plot 
## -- 2022-02-02  1.0.0     WB       Release of first version
## -- 2022-02-02  1.0.1     MRD      Cleaning the code
## -- 2022-02-10  1.0.2     WB       Introduce transparency in arrow depending on applied torque
## -- 2022-02-10  1.0.3     WB       Set init_angles as presets for starting angles
## -- 2022-02-10  1.0.4     WB       Normalize angle in reward calculation
## -- 2022-02-10  1.0.5     WB       Fix arrow head 
## -- 2022-02-14  1.0.6     WB       Update _compute_reward method
## -- 2022-02-17  1.0.7     WB       Taking into account the outer pole in reward calculation
## -- 2022-02-21  1.0.8     WB       Edit the formulation the of _compute_reward method
## -- 2022-03-02  1.0.9     WB       Include Torque and Change of state in _compute_reward method
## -- 2022-04-08  1.1.0     SY       Refactoring due to auto generated ID in class Dimension
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.1.0 (2022-04-08)

This module provides an RL environment of double pendulum.
"""

from mlpro.rl.models import *
from mlpro.bf.various import *
import numpy as np
from numpy import sin, cos
import matplotlib.pyplot as plt
from matplotlib.patches import Arc, RegularPolygon
import scipy.integrate as integrate
from collections import deque


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class DoublePendulum(Environment):
    """
    This is the main class of the Double Pendulum environment that inherits 
    Environment class from MLPro.
    
    Parameters
    ----------
    p_logging : Log, optional
        Logging functionalities. The default is Log.C_LOG_ALL.
    t_step : float, optional
        Time for each time step (in seconds). The default is 0.0025.
    t_act : int, optional
        Action frequency (with respect to the time step). The default is 20.
    max_torque : float, optional
        Maximum torque applied to pendulum 1. The default is 20.
    max_speed : float, optional
        Maximum speed applied to pendulum 1. The default is 10.
    l1 : float, optional
        Length of pendulum 1 in m. The default is 0.5
    l2 : float, optional
        Length of pendulum 2 in m. The default is 0.25
    m1 : float, optional
        Mass of pendulum 1 in kg. The default is 0.5
    m2 : float, optional
        Mass of pendulum 2 in kg. The default is 0.25
    init_angles: str, optional
        'up' starts the pendulum in an upright position
        'down' starts the pendulum in a downward position
        'random' starts the pendulum from a random position.
    g : float, optional
        Gravitational acceleration. The default is 9.8
    history_length : int, optional
        Historical trajectory points to display. The default is 2.
        
    Attributes
    ----------
    C_NAME : str
        Name of the environment.
    C_CYCLE_LIMIT : int
        The number of cycle limit.
    C_LATENCY : timedelta()
        Latency.
    C_REWARD_TYPE : Reward
        Rewarding type.
    """
    C_NAME = "DoublePendulum"
    C_CYCLE_LIMIT = 0
    C_LATENCY = timedelta(0, 0, 0)
    C_REWARD_TYPE = Reward.C_TYPE_OVERALL

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_logging=Log.C_LOG_ALL, t_step=0.0025, t_act=20, max_torque=20,
                 max_speed=10, l1=0.5, l2=0.25, m1=0.5, m2=0.25, init_angles='down', 
                 g=9.8, history_length=3):
        self.t_step = t_step
        self.t_act = t_act

        self.set_latency(timedelta(0, t_act * t_step, 0))

        self.max_torque = max_torque
        self.max_speed = max_speed

        self.l1 = l1
        self.l2 = l2
        self.L = l1 + l2
        self.m1 = m1
        self.m2 = m2
        self.M = m1+m2
        self.g = g
        
        self.th1dot = 0
        self.th2dot = 0
        
        if init_angles=='up':
            self.th1 = 180
            self.th2 = 180
        elif init_angles=='down':
            self.th1 = 0
            self.th2 = 0
        elif init_angles=='random':
            self.th1 = np.random.rand(1)[0]*180
            self.th2 = np.random.rand(1)[0]*180
        else:
            raise NotImplementedError("init_angles value must be up, down, or random")
        
        self.history_x = deque(maxlen=history_length)
        self.history_y = deque(maxlen=history_length)

        super().__init__(p_mode=Environment.C_MODE_SIM, p_logging=p_logging)

        self.C_SCIREF_TYPE = self.C_SCIREF_TYPE_ONLINE
        self.C_SCIREF_AUTHOR = "John Hunter, Darren Dale, Eric Firing, Michael \
                                Droettboom and the Matplotlib development team"
        self.C_SCIREF_TITLE = "The Double Pendulum Problem"
        self.C_SCIREF_URL = "https://matplotlib.org/stable/gallery/animation/double_pendulum.html"

        self._state = State(self._state_space)

        self.reset()

    ## -------------------------------------------------------------------------------------------------
    def setup_spaces(self):
        """
        This method is used to set up action and state spaces of the system.

        Returns
        -------
        state_space : ESpace()
            State space of the system.
        action_space : ESpace()
            Action space of the system.

        """
        state_space = ESpace()
        action_space = ESpace()

        state_space.add_dim(Dimension('theta 1', 'th1', 'Angle of Pendulum 1', '', 'degrees',
                                      '\textdegrees', [-np.pi, np.pi]))
        state_space.add_dim(Dimension('omega 1', 'w1', 'Angular Velocity of Pendulum 1', '',
                                      'degrees/second', '\textdegrees/s', [-np.inf, np.inf]))
        state_space.add_dim(Dimension('theta 2', 'th2', 'Angle of pendulum 2', '', 'degrees',
                                      '\textdegrees', [-np.pi, np.pi]))
        state_space.add_dim(Dimension('omega 2', 'w2', 'Angular Velocity of Pendulum 2', '',
                                      'degrees/second', '\textdegrees/s', [-np.inf, np.inf]))

        action_space.add_dim(Dimension('torque 1', 'tau1', 'Applied Torque of Motor 1', '',
                                       'Nm', 'Nm', [-self.max_torque, self.max_torque]))

        return state_space, action_space

    ## -------------------------------------------------------------------------------------------------
    def derivs(self, state, t=None):
        """
        This method is used to calculate the derivatives of the system, given the
        current states.

        Parameters
        ----------
        state : list
            [th, th1dot, th2, th2dot]

        t : list
            Timestep

        Returns
        -------
        dydx : list
            The derivatives of the given state

        """
        dydx = np.zeros_like(state)
        dydx[0] = state[1]

        delta = state[2] - state[0]
        den1 = (self.m1 + self.m2) * self.l1 - self.m2 * self.l1 * cos(delta) * cos(delta)
        dydx[1] = ((self.m2 * self.l1 * state[1] * state[1] * sin(delta) * cos(delta)
                    + self.m2 * self.g * sin(state[2]) * cos(delta)
                    + self.m2 * self.l2 * state[3] * state[3] * sin(delta)
                    - (self.m1 + self.m2) * self.g * sin(state[0]))
                   / den1)

        dydx[2] = state[3]

        den2 = (self.l2 / self.l1) * den1
        dydx[3] = ((- self.m2 * self.l2 * state[3] * state[3] * sin(delta) * cos(delta)
                    + (self.m1 + self.m2) * self.g * sin(state[0]) * cos(delta)
                    - (self.m1 + self.m2) * self.l1 * state[1] * state[1] * sin(delta)
                    - (self.m1 + self.m2) * self.g * sin(state[2]))
                   / den2)

        return dydx

    ## -------------------------------------------------------------------------------------------------
    @staticmethod
    def angle_normalize(x):
        """
        This method is called to ensure a normalized angle in radians.

        Returns
        -------
        angle : float
            Normalized angle.

        """
        return ((x + np.pi) % (2 * np.pi)) - np.pi

    ## -------------------------------------------------------------------------------------------------
    def _reset(self, p_seed=None) -> None:
        """
        This method is used to reset the environment.

        Parameters
        ----------
        p_seed : int, optional
            Not yet implemented. The default is None.

        """
        state_ids = self._state.get_dim_ids()
        self._state.set_value(state_ids[0], np.radians(self.th1))
        self._state.set_value(state_ids[1], np.radians(self.th1dot))
        self._state.set_value(state_ids[2], np.radians(self.th2))
        self._state.set_value(state_ids[3], np.radians(self.th2dot))

        self.history_x.clear()
        self.history_y.clear()
        self.action_cw = False
        self.alpha = 0

    ## -------------------------------------------------------------------------------------------------
    def _simulate_reaction(self, p_state: State, p_action: Action) -> State:
        """
        This method is used to calculate the next states of the system after a set of actions.

        Parameters
        ----------
        p_state : State
            State.
        p_action : Action
            Action.

        Returns
        -------
        _state : State
            Current states.

        """
        state = p_state.get_values()
        th1, th1dot, th2, th2dot = state
        torque = p_action.get_sorted_values()[0]
        torque = np.clip(torque, -self.max_torque, self.max_torque)
        
        self.alpha = abs(torque)/self.max_torque
        
        state[1] = th1dot + (3 * self.g / (2 * self.l1) * sin(th1) + 3.0 /
                             (self.m1 * self.l1 ** 2) * torque) * self.t_step

        if abs(th1dot) > self.max_speed:
            state[1] = np.clip(state[1], -th1dot, th1dot)
        else:
            state[1] = np.clip(state[1], -self.max_speed, self.max_speed)

        self.y = integrate.odeint(self.derivs, state, np.arange(0, self.t_act * self.t_step, self.t_step))
        state = self.y[-1]
        state[0] = DoublePendulum.angle_normalize(state[0])
        state[2] = DoublePendulum.angle_normalize(state[2])
        
        self.action_cw = True if torque <= 0 else False
        state_ids = self._state.get_dim_ids()
        for i in range(len(state)):
            self._state.set_value(state_ids[i], state[i])

        return self._state

    ## -------------------------------------------------------------------------------------------------
    def _compute_broken(self, p_state: State) -> bool:
        """
        This method computes the broken flag. This method can be redefined.

        Parameters
        ----------
        p_state : State
            State.

        Returns
        -------
        bool :
            Broken or not.

        """

        return False

    ## -------------------------------------------------------------------------------------------------
    def _compute_success(self, p_state: State) -> bool:
        """
        This method computes the success flag. This method can be redefined.

        Parameters
        ----------
        p_state : State
            State.

        Returns
        -------
        bool :
            Success or not success.

        """

        return False

    ## -------------------------------------------------------------------------------------------------
    def _compute_reward(self, p_state_old: State, p_state_new: State) -> Reward:
        """
        This method calculates the reward for C_TYPE_OVERALL reward type.

        Parameters
        ----------
        p_state_old : State
            Previous state.
        p_state_new : State
            New state.

        Returns
        -------
        reward : Reward
            Reward values.

        """
        reward = Reward(Reward.C_TYPE_OVERALL)
        
        target = np.array([np.pi, 0.0, np.pi, 0.0])
        state = p_state_new.get_values()
        old_state = p_state_old.get_values()
        
        th1_count = 0
        for th1 in self.y[::-1, 0]:
            ang = np.degrees(DoublePendulum.angle_normalize(th1))
            if ang > 170 or ang < 190 or \
                    ang < -170 or ang > -190:
                th1_count += 1
            else:
                break
        th1_distance = np.pi - abs(DoublePendulum.angle_normalize(np.radians(state[0])))
        th1_distance_costs = 4 if th1_distance <= 0.1 else 0.3 / th1_distance
        
        th1_speed_costs = np.pi * abs(state[1]) / self.max_speed
        
        # max acceleration in one timestep is assumed to be double the max speed
        th1_acceleration_costs = np.pi * abs(self.y[-1, 1]-self.y[-2, 1]) / (2 * self.max_speed)
        
        inner_pole_costs = (th1_distance_costs * th1_count / len(self.y)) - th1_speed_costs - (th1_acceleration_costs ** 0.5)
        inner_pole_weight = (self.l1/2)*self.m1
        
        th2_count = 0
        for th2 in self.y[::-1, 2]:
            ang = np.degrees(DoublePendulum.angle_normalize(th2))
            if ang > 170 or ang < 190 or \
                    ang < -170 or ang > -190:
                th2_count += 1
            else:
                break
        th2_distance = np.pi - abs(DoublePendulum.angle_normalize(np.radians(state[2])))
        th2_distance_costs = 4 if th2_distance <= 0.1 else 0.3 / th2_distance
        
        th2_speed_costs = np.pi * abs(state[3]) / self.max_speed
        
        th2_acceleration_costs = np.pi * abs(self.y[-1, 3]-self.y[-2, 3]) / (2 * self.max_speed)
        
        outer_pole_costs = (th2_distance_costs * th2_count / len(self.y)) - th2_speed_costs - (th2_acceleration_costs ** 0.5)
        outer_pole_weight = 0.5 * (self.l2/2)*self.m2
        
        change_costs = ((np.linalg.norm(target[::2] - np.array(old_state)[::2])*inner_pole_weight) - 
                        (np.linalg.norm(target[::2] - np.array(state)[::2])*outer_pole_weight))
        
        reward.set_overall_reward((inner_pole_costs * inner_pole_weight) + (outer_pole_costs * outer_pole_weight) )
                                  # - (self.alpha * np.pi/2) + (change_costs))

        return reward

    ## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        """
        This method initializes the plot figure of each episode. When the environment
        is reset, the previous figure is closed and reinitialized. 
        
        Parameters
        ----------
        p_figure : matplotlib.figure.Figure
            A Figure object of the matplotlib library.
        """
        if hasattr(self, 'fig'):
            plt.close(self.fig)

        if p_figure is None:
            self.fig = plt.figure(figsize=(5, 4))
            self.embedded_fig = False
        else:
            self.fig = p_figure
            self.embedded_fig = True

        self.ax = self.fig.add_subplot(autoscale_on=False,
                                       xlim=(-self.L * 1.2, self.L * 1.2), ylim=(-self.L * 1.2, self.L * 1.2))
        self.ax.set_aspect('equal')
        self.ax.grid()

        self.cw_arc = Arc([0, 0], 0.5 * self.l1, 0.5 * self.l1, angle=0, theta1=0,
                          theta2=250, color='crimson')
        endX = (0.5 * self.l1 / 2) * np.cos(np.radians(0))
        endY = (0.5 * self.l1 / 2) * np.sin(np.radians(0))
        self.cw_arrow = RegularPolygon((endX, endY), 3, 0.5 * self.l1 / 9, np.radians(180),
                                       color='crimson')

        self.ccw_arc = Arc([0, 0], 0.5 * self.l1, 0.5 * self.l1, angle=70, theta1=0,
                           theta2=320, color='crimson')
        endX = (0.5 * self.l1 / 2) * np.cos(np.radians(70 + 320))
        endY = (0.5 * self.l1 / 2) * np.sin(np.radians(70 + 320))
        self.ccw_arrow = RegularPolygon((endX, endY), 3, 0.5 * self.l1 / 9, np.radians(70 + 320),
                                        color='crimson')

        self.ax.add_patch(self.cw_arc)
        self.ax.add_patch(self.cw_arrow)
        self.ax.add_patch(self.ccw_arc)
        self.ax.add_patch(self.ccw_arrow)

        self.cw_arc.set_visible(False)
        self.cw_arrow.set_visible(False)
        self.ccw_arc.set_visible(False)
        self.ccw_arrow.set_visible(False)

        self.line, = self.ax.plot([], [], 'o-', lw=2)
        self.trace, = self.ax.plot([], [], '.-', lw=1, ms=2)

    ## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        """
        This method updates the plot figure of each episode. When the figure is
        detected to be an embedded figure, this method will only set up the 
        necessary data of the figure.
        
        """
        x1 = self.l1 * sin(self.y[:, 0])
        y1 = -self.l1 * cos(self.y[:, 0])

        x2 = self.l2 * sin(self.y[:, 2]) + x1
        y2 = -self.l2 * cos(self.y[:, 2]) + y1

        thisx = [0, x1[-1], x2[-1]]
        thisy = [0, y1[-1], y2[-1]]
        self.history_x.appendleft(thisx[2])
        self.history_y.appendleft(thisy[2])
        self.line.set_data(thisx, thisy)
        self.trace.set_data(self.history_x, self.history_y)

        if self.action_cw:
            self.cw_arc.set_visible(True)
            self.cw_arrow.set_visible(True)
            self.cw_arc.set_alpha(self.alpha)
            self.cw_arrow.set_alpha(self.alpha)
            self.ccw_arc.set_visible(False)
            self.ccw_arrow.set_visible(False)
        else:
            self.cw_arc.set_visible(False)
            self.cw_arrow.set_visible(False)
            self.ccw_arc.set_visible(True)
            self.ccw_arrow.set_visible(True)
            self.ccw_arc.set_alpha(self.alpha)
            self.ccw_arrow.set_alpha(self.alpha)

        if not self.embedded_fig:
            self.fig.canvas.draw()
            self.fig.canvas.flush_events()



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.rl.envmodels
## -- Module  : mlp_robotinhtm
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2022-05-20  0.0.0     MRD      Creation
## -- 2022-05-22  1.0.0     MRD      Release first version
## -- 2022-05-30  1.0.1     MRD      Cleaning up HTMEnvModel, now inherit directly from the
## --                                actual environment
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.1 (2022-05-30)

This module provides Environment Model based on Homogeneous Transformations Matrix 
Neural Network for robotinhtm environment.
"""

import torch
import transformations

from mlpro.rl.models import *
from mlpro.rl.pool.envs.robotinhtm import RobotArm3D
from mlpro.rl.pool.envs.robotinhtm import RobotHTM
from mlpro.sl.pool.afct.afct_pytorch import TorchAFct

from torch.utils.data.sampler import SubsetRandomSampler
from collections import deque

class DHLayer(torch.nn.Module):
    """
    This class represents a layer architecture based on DH Parameter as the learnable parameter
    and use them to construct the transformation matrix
    :param p_in: Number of Joints
    """

    def __init__(self, p_in):
        super(DHLayer, self).__init__()
        self._in = p_in

        self.register_parameter("alpha", torch.nn.Parameter((torch.rand(self._in, 1) - 0.5) * 1))
        self.register_parameter("beta", torch.nn.Parameter((torch.rand(self._in, 1) - 0.5) * 1))
        self.register_parameter("a", torch.nn.Parameter((torch.rand(self._in, 1) - 0.5) * 1))
        self.register_parameter("b", torch.nn.Parameter((torch.rand(self._in, 1) - 0.5) * 1))
        self.register_parameter("d", torch.nn.Parameter((torch.rand(self._in, 1) - 0.5) * 1))

    def forward(self, p_in):
        batch_size = p_in.shape[0]

        l_alpha = self.alpha.repeat(batch_size, 1).reshape(batch_size, self._in)
        l_beta = self.beta.repeat(batch_size, 1).reshape(batch_size, self._in)
        l_tx = self.a.repeat(batch_size, 1, 1).reshape(batch_size, self._in, 1, 1)
        l_ty = self.b.repeat(batch_size, 1, 1).reshape(batch_size, self._in, 1, 1)
        l_tz = self.d.repeat(batch_size, 1, 1).reshape(batch_size, self._in, 1, 1)

        # Construct Z Transformations
        unit = torch.Tensor([[0.0, 0.0, 1.0]], device=p_in.device)
        trans_z = self.construct_transformation(unit, p_in, l_tz)

        # Construct Y Transformations
        unit = torch.Tensor([[0.0, 1.0, 0.0]])
        trans_y = self.construct_transformation(unit, l_beta, l_ty)

        # Construct X Transformations
        unit = torch.Tensor([[1.0, 0.0, 0.0]], device=p_in.device)
        trans_x = self.construct_transformation(unit, l_alpha, l_tx)

        # Construct DH Matrix
        # dh_mat = trans_z * trans_y * trans_x
        dh_mat = torch.matmul(trans_z, trans_y)
        dh_mat = torch.matmul(dh_mat, trans_x)

        return dh_mat

    def construct_transformation(self, p_unit, p_angle, p_transl):
        """
        Construct Transformation Matrix
        :param p_unit: Rotation and Translation Unit Vector
        :param p_angle: Rotation Angle
        :param p_transl: Translation
        :return: trans_mat: Transformations Matrix
        """

        # Create Rotation Matrix
        rot_mat = self.create_rot_mat(p_unit.to(p_angle.device), p_angle)

        # Create Translation Vector
        transl_vec = p_unit * p_transl

        # Combine Rotation Matrix and Translation Vector into Transformation Matrix
        trans_mat = self.trans_rot_to_transformations_mat(transl_vec, rot_mat)

        return trans_mat

    def trans_rot_to_transformations_mat(self, p_trans, p_rot):
        """
        Combine Translation Vector and Rotation Matrix into Transformation Matrix
        :param p_trans: Translation Vector
        :param p_rot: Rotation Matrix
        :return: trans_mat: Transformation Matrix
        """
        batch_size = p_trans.shape[0]
        fixed_vec = torch.Tensor([[0.0, 0.0, 0.0, 1.0]], device=p_trans.device)
        trans_mat = torch.cat([p_rot, p_trans.permute(0, 1, 3, 2)], dim=3)
        trans_mat = torch.cat([trans_mat, fixed_vec.repeat(batch_size, self._in, 1, 1)], dim=2)
        return trans_mat

    def create_trans_vec(self):
        pass

    def create_rot_mat(self, p_unit, p_angle):
        """
        Create Rotation Matrix
        :param p_unit: Rotation Unit Vector
        :param p_angle: Rotation Angle
        :return: rot_mat: Rotation Matrix
        """
        batch_size = p_angle.shape[0]
        masking_indices = torch.tensor([[3, 2, 1], [2, 3, 0], [1, 0, 3]], device=p_angle.device)
        masking_cross = torch.Tensor([[0, -1, 1], [1, 0, -1], [-1, 1, 0]], device=p_angle.device)

        unit = p_unit
        unit_stack = torch.Tensor([], device=p_angle.device)
        outer_prod = torch.Tensor([], device=p_angle.device)

        for i in range(self._in):
            outer_prod = torch.cat([outer_prod, torch.ger(unit[0], unit[0])])
            k_unit = torch.cat([unit, torch.Tensor([[0.0]], device=p_angle.device)], dim=1)
            unit_store = k_unit[0][masking_indices] * masking_cross
            unit_stack = torch.cat([unit_stack, unit_store])

        unit = unit_stack.reshape(self._in, 3, 3).repeat(batch_size, 1, 1, 1)
        outer = outer_prod.reshape(self._in, 3, 3).repeat(batch_size, 1, 1, 1)

        rot_mat = unit * torch.sin(p_angle).reshape(batch_size, self._in, 1, 1) + (
                torch.eye(3, device=p_angle.device).repeat(batch_size, 1, 1, 1) - outer) * \
                  torch.cos(p_angle).reshape(batch_size, self._in, 1, 1) + outer

        return rot_mat

class N3(torch.nn.Module):
    def __init__(self, n_in):
        super(N3, self).__init__()
        self.n_in = n_in
        self.added = 0

    def forward(self, I):
        BatchSize = I.shape[0]

        A = torch.eye(4)
        b = torch.tril(torch.ones(self.n_in + self.added + 1, self.n_in + self.added)).flatten()
        c = torch.triu(torch.ones(self.n_in + self.added + 1, self.n_in + self.added), diagonal=1).flatten()

        maskIlower = torch.einsum('ij,k->kij', A, b).reshape(self.n_in + self.added + 1, self.n_in + self.added, 4, 4)
        maskIupper = torch.einsum('ij,k->kij', A, c).reshape(self.n_in + self.added + 1, self.n_in + self.added, 4, 4)

        output1 = torch.matmul(maskIlower, I.reshape(BatchSize, 1, self.n_in + self.added, 4, 4))
        output1 = torch.add(output1, maskIupper)
        output1 = output1.permute(0, 2, 1, 3, 4)

        output = torch.eye(4).repeat(BatchSize, self.n_in + self.added + 1, 1, 1)
        for outnum in range(self.n_in + self.added):
            output = torch.matmul(output, output1[:, outnum])

        return output

class RobotHTMModel(torch.nn.Module):
    """
    Provide Forward Kinematic based on Neural Network with DH Layer.
    Predicts the end-effector position with given joint angles.
    """

    def __init__(self, p_in, p_t):
        super(RobotHTMModel, self).__init__()
        self._in = p_in
        self._t = p_t

        self.dh_layer = DHLayer(self._in)
        self.eef_layer = N3(self._in)

    def forward(self, p_in):
        batch_size=p_in.shape[0]
        new_i = p_in.reshape(batch_size,2,self._in) * torch.cat([torch.Tensor([self._t]).repeat(1,self._in), torch.ones(1,self._in)])
        new_i = torch.sum(new_i,dim=1)
        
        trans = self.dh_layer(new_i)
        output = self.eef_layer(trans)
        
        return output

class IOElement(BufferElement):
    def __init__(self, p_input: torch.Tensor, p_output: torch.Tensor):

        super().__init__({"input": p_input, "output": p_output})


# Buffer
class MyOwnBuffer(Buffer, torch.utils.data.Dataset):
    def __init__(self, p_size=1):
        Buffer.__init__(self, p_size=p_size)
        self._internal_counter = 0

    def add_element(self, p_elem: BufferElement):
        Buffer.add_element(self, p_elem)
        self._internal_counter += 1

    def get_internal_counter(self):
        return self._internal_counter

    def __getitem__(self,idx):
        return self._data_buffer["input"][idx], self._data_buffer["output"][idx]

class RobothtmAFct(TorchAFct):
    C_NAME = "Robothtm Adaptive Function"
    C_BUFFER_CLS = MyOwnBuffer

    def _setup_model(self):
        self.joint_num = self._output_space.get_num_dim() - 6
        self.net_model = RobotHTMModel(self.joint_num, 0.01)
        self.optimizer = torch.optim.Adam(self.net_model.parameters(), lr=3e-4)
        self.loss_dyn = torch.nn.MSELoss()
        self.train_model = True
        self.input_temp = None

        self.sim_env = RobotArm3D()

        joints = []

        jointType = []
        vectLinkLength = [[0, 0, 0], [0, 0, 0]]
        jointType.append("rz")
        for joint in range(self.joint_num - 1):
            vectLinkLength.append([0, 0.7, 0])
            jointType.append("rx")

        jointType.append("f")

        for x in range(len(jointType)):
            vectorLink = dict(x=vectLinkLength[x][0], y=vectLinkLength[x][1], z=vectLinkLength[x][2])
            joint = dict(
                Joint_name="Joint %d" % x,
                Joint_type=jointType[x],
                Vector_link_length=vectorLink,
            )
            joints.append(joint)

        for robo in joints:
            self.sim_env.add_link_joint(
                lvector=torch.Tensor(
                    [
                        [
                            robo["Vector_link_length"]["x"],
                            robo["Vector_link_length"]["y"],
                            robo["Vector_link_length"]["z"],
                        ]
                    ]
                ),
                jointAxis=robo["Joint_type"],
                thetaInit=torch.Tensor([np.radians(0)]),
            )

        self.sim_env.update_joint_coords()

    def _input_preproc(self, p_input: torch.Tensor) -> torch.Tensor:
        input = torch.cat([p_input[0][6+self.joint_num:], p_input[0][6:6+self.joint_num]])
        input = input.reshape(1,self.joint_num*2)
        self.input_temp = p_input[0][:3].reshape(1,3)
        
        return input

    def _output_postproc(self, p_output: torch.Tensor) -> torch.Tensor:
        angles = torch.Tensor([])
        thets = torch.zeros(3)
        for idx in range(self.joint_num):
            angle = torch.Tensor(transformations.euler_from_matrix(p_output[-1][idx][:].detach().numpy(), 'rxyz')) - thets
            thets = torch.Tensor(transformations.euler_from_matrix(p_output[-1][idx][:].detach().numpy(), 'rxyz'))
            angles = torch.cat([angles, torch.norm(angle).reshape(1, 1)], dim=1)

        output = torch.cat([self.input_temp, p_output[-1][-1][:3, [-1]].reshape(1,3)], dim=1)
        output = torch.cat([output, angles], dim=1)

        return output
    
    def _adapt(self, p_input: Element, p_output: Element) -> bool:
        model_input = deque(p_input.get_values()[6:])
        model_input.rotate(self.joint_num)
        model_input = torch.Tensor([list(model_input)])

        self.sim_env.set_theta(torch.Tensor([p_output.get_values()[6 : 6 + self.joint_num]]))
        self.sim_env.update_joint_coords()

        model_output = self.sim_env.get_homogeneous().reshape(self.joint_num+1,4,4)

        self._add_buffer(IOElement(model_input, model_output))

        if self._buffer.get_internal_counter() % 100 != 0:
            return False

        # Divide Test and Train
        if self.train_model:
            dataset_size = len(self._buffer)
            indices = list(range(dataset_size))
            split = int(np.floor(0.2 * dataset_size))
            np.random.seed(random.randint(1,1000))
            np.random.shuffle(indices)
            train_indices, test_indices = indices[split:], indices[:split]

            train_sampler = SubsetRandomSampler(train_indices)
            test_sampler = SubsetRandomSampler(test_indices)
            trainer = torch.utils.data.DataLoader(self._buffer, batch_size=100, sampler=train_sampler)
            tester = torch.utils.data.DataLoader(self._buffer, batch_size=100, sampler=test_sampler)

            # Training
            self.net_model.train()

            for i, (In, Label) in enumerate(trainer):
                outputs = self.net_model(In)
                loss = self.loss_dyn(outputs, Label)
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

            test_loss = 0

            self.net_model.eval()
            for i, (In, Label) in enumerate(tester):
                outputs = self.net_model(In)
                loss = self.loss_dyn(outputs, Label)
                test_loss += loss.item()


            if test_loss/len(tester) < 5e-9:
                self.train_model = False

        return True

    def _add_buffer(self, p_buffer_element: IOElement):
        self._buffer.add_element(p_buffer_element)

class HTMEnvModel(RobotHTM, EnvModel):
    C_NAME = "HTM Env Model"

    def __init__(
        self,
        p_num_joints=4,
        p_target_mode="Random",
        p_ada=True,
        p_logging=False,
    ):

        RobotHTM.__init__(self, p_num_joints=p_num_joints, p_target_mode=p_target_mode)
        
        # Setup Adaptive Function
        # HTM Function Here
        afct_strans = AFctSTrans(
            RobothtmAFct,
            p_state_space=self._state_space,
            p_action_space=self._action_space,
            p_threshold=1.8,
            p_buffer_size=20000,
            p_ada=p_ada,
            p_logging=p_logging,
        )

        EnvModel.__init__(
            self,
            p_observation_space=self._state_space,
            p_action_space=self._action_space,
            p_latency=timedelta(seconds=self.dt),
            p_afct_strans=afct_strans,
            p_afct_reward=None,
            p_afct_success=None,
            p_afct_broken=None,
            p_ada=p_ada,
            p_logging=p_logging,
        )

        self.reset()




## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro.bf
## -- Module  : data
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-04-16  0.0.0     DA       Creation
## -- 2021-05-29  1.0.0     DA       Release of first version
## -- 2021-06-16  1.1.0     SY       Adding the first version of data storing,
## --                                data plotting, and data saving classes
## -- 2021-06-21  1.2.0     SY       Add extensions in classes Loadable,
## --                                Saveable, DataPlotting & DataStoring.
## -- 2021-08-28  1.2.1     DA       Added constant C_VAR0 to class DataStoring
## -- 2021-09-18  1.2.1     MRD      Buffer Class Implementation. Add new parameter buffer
## --                                to the Adaptive Class
## -- 2021-09-19  1.2.2     MRD      Improvement on Buffer Class. Implement new base class
## --                                Buffer Element and BufferRnd
## -- 2021-09-22  1.3.0     MRD      New classes BufferElement, Buffer, BufferRnd
## -- 2021-09-25  1.3.1     MRD      Add __len__ functionality for SARBuffer
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.3.1 (2021-09-25)

This module provides various elementary data management classes.
"""

from datetime import datetime, timedelta
# from time import sleep
import numpy as np
import math
import matplotlib.pyplot as plt
import pickle as pkl
import os
import csv
import copy
from mlpro.bf.various import LoadSave
import random


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class DataStoring(LoadSave):
    """
    This class provides a functionality to store values of variables during
    training/simulation.
    """

    C_VAR0 = 'Frame ID'

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_variables):
        """
        Parameters:
            p_variable    List of variable names
        """
        self.memory_dict = {}
        self.names = p_variables
        self.frame_id = {}
        for name in self.names:
            self.memory_dict[name] = {}
            self.frame_id[name] = []

    ## -------------------------------------------------------------------------------------------------
    def add_frame(self, p_frame_id):
        """
        To store unique sections in a variable (e.g episodes in RL, etc.)
        """
        for name in self.names:
            self.memory_dict[name][p_frame_id] = []
            self.frame_id[name].append(p_frame_id)

    ## -------------------------------------------------------------------------------------------------
    def memorize(self, p_variable, p_frame_id, p_value):
        """
        To store a particular variable into a memory
        """
        self.memory_dict[p_variable][p_frame_id].append(p_value)

    ## -------------------------------------------------------------------------------------------------
    def get_values(self, p_variable, p_frame_id=None):
        """
        To obtain value from the memory
        """
        if p_frame_id == None:
            return self.memory_dict[p_variable]
        else:
            return self.memory_dict[p_variable][p_frame_id]

    ## -------------------------------------------------------------------------------------------------
    def list_to_chunks(self, p_data, p_chunksize):
        NumChunks = int(math.ceil(len(p_data) / (p_chunksize * 1.0)))
        retval = []
        for chunk in range(NumChunks):
            retval.append(sum(p_data[chunk * p_chunksize: (chunk + 1) * p_chunksize]) / (1.0 * p_chunksize))
        return retval

    ## -------------------------------------------------------------------------------------------------
    def compress(self, p_chunksize):
        for name in self.names:
            for ep in len(self.memory_dict[name]):
                self.memory_dict[name][ep] = self.list_to_chunks(self.memory_dict[name][ep], p_chunksize)

    ## -------------------------------------------------------------------------------------------------
    def save_data(self, p_path, p_filename=None, p_delimiter="\t") -> bool:
        """
        To save stored data in memory_dict as a readable file format
        """

        if (p_filename is not None) and (p_filename != ''):
            self.filename = p_filename
        else:
            self.filename = self.generate_filename()

        if self.filename is None:
            return False

        try:
            if not os.path.exists(p_path):
                os.makedirs(p_path)
            path_save = p_path + os.sep + self.filename + ".csv"
            with open(path_save, "w", newline="") as write_file:
                header = copy.deepcopy(self.names)
                header.insert(0, self.C_VAR0)
                writer = csv.writer(write_file, delimiter=p_delimiter, quoting=csv.QUOTE_ALL)
                writer.writerow(header)
                writer = csv.writer(write_file, delimiter=p_delimiter)
                for frame in self.frame_id[self.names[0]]:
                    for idx in range(len(self.memory_dict[self.names[0]][frame])):
                        row = []
                        row.append(frame)
                        for name in self.names:
                            row.append(self.memory_dict[name][frame][idx])
                        writer.writerow(row)
            return True
        except:
            return False

    ## -------------------------------------------------------------------------------------------------
    def load_data(self, p_path, p_filename, p_delimiter="\t") -> bool:
        """
        To load data from a readable file format and store them into the DataStoring class format
        """

        try:
            path_load = p_path + os.sep + p_filename
            with open(path_load, "r") as read_file:
                reader = csv.reader(read_file, delimiter=p_delimiter)
                header = True
                for row in reader:
                    if header:
                        del row[0:1]
                        self.__init__(row)
                        header = False
                    else:
                        column = 1
                        for name in self.names:
                            if row[0] not in self.frame_id[name]:
                                self.add_frame(row[0])
                            self.memorize(name, row[0], float(row[column]))
                            column += 1
            return True
        except:
            return False


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class BufferElement:
    """
    Base class implementation for buffer element
    """

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_element: dict) -> None:
        """
        Parameters:
            p_element (dict): Buffer element in dictionary
        """

        self._element = {}

        self.add_value_element(p_element)

    ## -------------------------------------------------------------------------------------------------
    def add_value_element(self, p_val: dict):
        """
        Adding new value to the element container

        Parameters:
            p_val (dict): Elements in dictionary
        """
        self._element = {**self._element, **p_val}

    ## -------------------------------------------------------------------------------------------------
    def get_data(self):
        """
        Get the buffer element.

        Returns:
            Returns the buffer element.
        """

        return self._element


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Buffer:
    """
    Base class implementation for buffer management.
    """

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_size=1):
        """
        Parameters:
            p_size (int, optional): Buffer size. Defaults to 1.
        """
        self._size = p_size
        self._data_buffer = {}

    ## -------------------------------------------------------------------------------------------------
    def add_element(self, p_elem: BufferElement):
        """
        Add element to the buffer.

        Parameters:
            p_elem (BufferElement): Element of Buffer
        """
        self._data_buffer = {**p_elem.get_data(), **self._data_buffer}
        for key, value in self._data_buffer.items():
            if key in p_elem.get_data() and key in self._data_buffer:
                if not isinstance(self._data_buffer[key], list):
                    self._data_buffer[key] = [p_elem.get_data()[key]]
                else:
                    self._data_buffer[key].append(p_elem.get_data()[key])

                if len(self._data_buffer[key]) > self._size:
                    self._data_buffer[key].pop(-len(self._data_buffer[key]))

    ## -------------------------------------------------------------------------------------------------
    def clear(self):
        """
        Resets buffer.
        """

        self._data_buffer.clear()

    ## -------------------------------------------------------------------------------------------------
    def get_latest(self):
        """
        Returns latest buffered element. 
        """

        try:
            return {key: self._data_buffer[key][-1] for key in self._data_buffer}
        except:
            return None

    ## -------------------------------------------------------------------------------------------------
    def get_all(self):
        """
        Return all buffered elements.

        """
        return self._data_buffer

    ## -------------------------------------------------------------------------------------------------
    def get_sample(self, p_num: int):
        """
        Sample some element from the buffer.

        Parameters:
            p_num (int): Number of sample

        Returns:
            Samples in dictionary
        """
        return self._extract_rows(self._gen_sample_ind(p_num))

    ## -------------------------------------------------------------------------------------------------
    def _gen_sample_ind(self, p_num: int) -> list:
        """
        Generate random indices from the buffer.

        Parameters:
            p_num (int): Number of sample

        Returns:
            List of incides
        """
        raise NotImplementedError

    ## -------------------------------------------------------------------------------------------------
    def _extract_rows(self, p_list_idx: list):
        """
        Extract the element in the buffer based on a
        list of indices.

        Parameters:
            p_list_idx (list): List of indices

        Returns:
            Samples in dictionary
        """
        rows = {}
        for key in self._data_buffer:
            rows[key] = [self._data_buffer[key][i] for i in p_list_idx]
        return rows

    ## -------------------------------------------------------------------------------------------------
    def is_full(self) -> bool:
        """
        Check if the buffer is full.

        Returns:
            True, if the buffer is full
        """
        keys = list(self._data_buffer.keys())
        return len(self._data_buffer[keys[0]]) >= self._size

    ## -------------------------------------------------------------------------------------------------
    def __len__(self):
        keys = list(self._data_buffer.keys())
        return len(self._data_buffer[keys[0]])


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class BufferRnd(Buffer):
    """
    Buffer implmentation with random sampling
    """

    ## -------------------------------------------------------------------------------------------------
    def _gen_sample_ind(self, p_num: int) -> list:
        """
        Generate random indicies

        Parameters:
            p_num (int): Number of sample

        Returns:
            List of indicies
        """
        keys = list(self._data_buffer.keys())
        return random.sample(list(range(0, len(self._data_buffer[keys[0]]))), p_num)



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_rl_006_run_own_agents_with_petting_zoo_environment.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-08-26  0.0.0     SY       Creation
## -- 2021-08-27  1.0.0     SY       Released first version
## -- 2021-09-11  1.0.0     MRD      Change Header information to match our new library name
## -- 2021-09-23  1.1.0     SY       Updated wrapper WrEnvPZoo class, provides two different envs
## -- 2021-09-29  1.1.1     SY       Change name: WrEnvPZoo to WrEnvPZOO2MLPro
## -- 2021-10-06  1.1.2     DA       Refactoring 
## -- 2021-10-18  1.1.3     DA       Refactoring 
## -- 2021-11-15  1.1.4     DA       Refactoring 
## -- 2021-11-15  1.1.4     DA       Refactoring 
## -- 2021-11-16  1.1.5     DA       Added explicit scenario reset with constant seeding 
## -- 2021-12-03  1.1.6     DA       Refactoring 
## -- 2022-02-25  1.1.7     SY       Refactoring due to auto generated ID in class Dimension
## -- 2022-05-19  1.1.8     SY       Utilize RandomGenerator
## -- 2022-05-30  1.1.9     DA       Cleaned up/rearranged a bit
## -- 2022-05-30  1.1.8     SY       Update pistonball_v5 to pistonball_v6
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.1.9 (2022-05-30)

This module shows how to run an own policy inside the MLPro standard agent model with a wrapped Petting Zoo environment.
"""


from pettingzoo.butterfly import pistonball_v6
from pettingzoo.classic import connect_four_v3
from mlpro.bf.math import *
from mlpro.rl.models import *
from mlpro.wrappers.pettingzoo import WrEnvPZOO2MLPro
import random
from mlpro.rl.pool.policies.randomgenerator import RandomGenerator



# 1 RL Scenario based on PettingZoo Pistonball environment
class PBScenario (RLScenario):
    """
    Reference : https://www.pettingzoo.ml/butterfly/pistonball
    """

    C_NAME      = 'Pistonball V6'

    def _setup(self, p_mode, p_ada, p_logging):
        zoo_env             = pistonball_v6.env()
        self._env           = WrEnvPZOO2MLPro(zoo_env, p_logging=p_logging)
        
        multi_agent         = MultiAgent(p_name='Pistonball_agents', p_ada=1, p_logging=True)
        agent_idx           = 0
        for k in self._env._zoo_env.action_spaces:
            agent_name      = "Agent_"+str(agent_idx)
            as_ids          = self._env.get_action_space().get_dim_ids()
            agent_ospace    = self._env.get_state_space()
            agent_asspace   = self._env.get_action_space().spawn([as_ids[agent_idx]])
            agent           = Agent(p_policy=RandomGenerator(p_observation_space=agent_ospace,
                                                             p_action_space=agent_asspace,
                                                             p_buffer_size=10,
                                                             p_ada=p_ada,
                                                             p_logging=p_logging
                                                             ),
                                    p_envmodel=None,
                                    p_id=agent_idx,
                                    p_name=agent_name,
                                    p_ada=p_ada,
                                    p_logging=p_logging
                                    )
            multi_agent.add_agent(p_agent=agent)
            agent_idx += 1

        return multi_agent
    
    

# 2 Alternative RL Scenario based on PettingZoo Connect Four environment
class C4Scenario (RLScenario):
    """
    https://www.pettingzoo.ml/classic/connect_four
    """

    C_NAME      = 'Connect Four V3'

    def _setup(self, p_mode, p_ada, p_logging):
        zoo_env             = connect_four_v3.env()
        self._env           = WrEnvPZOO2MLPro(zoo_env, p_logging=True)
        
        multi_agent         = MultiAgent(p_name='Connect4_Agents', p_ada=1, p_logging=True)
        agent_idx           = 0
        for k in self._env._zoo_env.action_spaces:
            agent_name      = "Agent_"+str(agent_idx)
            as_ids          = self._env.get_action_space().get_dim_ids()
            agent_sspace    = self._env.get_state_space()
            agent_asspace   = self._env.get_action_space().spawn([as_ids[agent_idx]])
            agent           = Agent(p_policy=RandomGenerator(p_observation_space=agent_sspace,
                                                             p_action_space=agent_asspace,
                                                             p_buffer_size=10,
                                                             p_ada=p_ada,
                                                             p_logging=p_logging
                                                             ),
                                    p_envmodel=None,
                                    p_id=agent_idx,
                                    p_name=agent_name,
                                    p_ada=p_ada,
                                    p_logging=p_logging
                                    )
            multi_agent.add_agent(p_agent=agent)
            agent_idx += 1

        return multi_agent

  

# 3 Create scenario and run some cycles
if __name__ == "__main__":
    # 3.1 Parameters for demo mode
    logging     = Log.C_LOG_ALL
    visualize   = True
  
else:
    # 3.2 Parameters for internal unit test
    logging     = Log.C_LOG_NOTHING
    visualize   = False
 

# 3.3 Instantiate one of two prepared demo scenarios
myscenario  = PBScenario(
        p_mode=Mode.C_MODE_SIM,
        p_ada=True,
        p_cycle_limit=100,
        p_visualize=visualize,
        p_logging=logging
)

# myscenario  = C4Scenario(
#         p_mode=Mode.C_MODE_SIM,
#         p_ada=True,
#         p_cycle_limit=100,
#         p_visualize=visualize,
#         p_logging=logging
# )


# 3.4 Reset and run the scenario
myscenario.reset(1)
myscenario.run() 


## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro
## -- Module  : test_environment
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-09-11  1.0.0     MRD      Creation
## -- 2021-09-11  1.0.0     MRD      Release First Version
## -- 2021-09-22  1.0.1     WB       Change Environment Instantiation Method
## -- 2021-09-26  1.0.2     MRD      Change the structure to work with GitHub Automated Test
## -- 2021 09-26  1.0.3     MRD      Change the import module due to the change of the pool
## --                                folder structer
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.3 (2021-09-26)

Unit test classes for environment.
"""


import pytest
import random
import numpy as np
from mlpro.rl.models import *
from mlpro.rl.pool.envs.robotinhtm import RobotHTM
from mlpro.rl.pool.envs.bglp import BGLP
from mlpro.rl.pool.envs.gridworld import GridWorld
from mlpro.rl.pool.envs.multicartpole import MultiCartPole
from mlpro.rl.pool.envs.doublependulum import DoublePendulum
# from mlpro.rl.pool.envs.ur5jointcontrol import UR5JointControl


## -------------------------------------------------------------------------------------------------
@pytest.mark.parametrize("env_cls", [RobotHTM, BGLP, GridWorld, MultiCartPole, DoublePendulum])
def test_environment(env_cls):
    env = env_cls()
    assert isinstance(env, Environment)
    
    assert isinstance(env.get_state_space(), ESpace)
    assert env.get_state_space().get_num_dim() != 0
    
    assert isinstance(env.get_action_space(), ESpace)
    assert env.get_action_space().get_num_dim() != 0
    
    state = env.get_state()
    
    assert isinstance(state, State)
        
    my_action_values = np.zeros(env.get_action_space().get_num_dim())
    for d in range(env.get_action_space().get_num_dim()):
        my_action_values[d] = random.random() 

    my_action_values = Action(0, env.get_action_space(), my_action_values)

    env.process_action(my_action_values)

    reward = env.compute_reward()
    
    assert isinstance(reward, Reward)

    env.reset()



from openai_ros import robot_gazebo_env


class MyRobotEnv(robot_gazebo_env.RobotGazeboEnv):
    """Superclass for all Robot environments.
    """

    def __init__(self):
        """Initializes a new Robot environment.
        """
        # Variables that we give through the constructor.

        # Internal Vars
        self.controllers_list = ['my_robot_controller1','my_robot_controller2', ..., 'my_robot_controllerX']

        self.robot_name_space = "my_robot_namespace"

        reset_controls_bool = True or False
        
        # We launch the init function of the Parent Class robot_gazebo_env.RobotGazeboEnv
        
        super(MyRobotEnv, self).__init__(controllers_list=self.controllers_list,
                                                robot_name_space=self.robot_name_space,
                                                reset_controls=reset_controls_bool)

    # Methods needed by the RobotGazeboEnv
    # ----------------------------
    
    

    def _check_all_systems_ready(self):
        """
        Checks that all the sensors, publishers and other simulation systems are
        operational.
        """
        # TODO
        return True
    
    # Methods that the TrainingEnvironment will need to define here as virtual
    # because they will be used in RobotGazeboEnv GrandParentClass and defined in the
    # TrainingEnvironment.
    # ----------------------------
    def _set_init_pose(self):
        """Sets the Robot in its init pose
        """
        raise NotImplementedError()
    
    
    def _init_env_variables(self):
        """Inits variables needed to be initialised each time we reset at the start
        of an episode.
        """
        raise NotImplementedError()

    def _compute_reward(self, observations, done):
        """Calculates the reward to give based on the observations given.
        """
        raise NotImplementedError()

    def _set_action(self, action):
        """Applies the given action to the simulation.
        """
        raise NotImplementedError()

    def _get_obs(self):
        raise NotImplementedError()

    def _is_done(self, observations):
        """Checks if episode done based on observations given.
        """
        raise NotImplementedError()
        
    # Methods that the TrainingEnvironment will need.
    # ----------------------------


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.rl.pool.envs.ur5jointcontrol
## -- Module  : urjointcontrol.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-09-13  0.0.0     WB       Creation
## -- 2021-09-13  1.0.0     WB       Released first version
## -- 2021-09-13  1.0.1     WB       Instantiated without WrEnvGym
## -- 2021-09-23  1.0.2     WB       Increased C_LATENCY
## -- 2021-10-05  1.0.3     SY       Update following new attributes done and broken in State
## -- 2021-12-03  1.0.4     DA       Refactoring
## -- 2021-12-20  1.0.5     DA       Replaced 'done' by 'success'
## -- 2021-12-20  1.0.6     WB       Update 'success' and 'broken' rule
## -- 2021-12-21  1.0.7     DA       Class UR5JointControl: renamed method reset() to _reset()
## -- 2022-01-20  1.0.8     MRD      Use the gym wrapper to wrap the ur5 environment
## -- 2022-03-04  1.1.0     WB       Adds the ability to control gripper
## -- 2022-06-06  2.0.0     MRD      Add ability to self build the ros workspace
## --                       MRD      Add the connection to the real robot, wrapped in Gym environment
## -- 2022-06-07  2.0.1     MRD      Define _export_action and _import_action for each communication type
## -- 2022-06-11  2.1.0     MRD      Finalizing simulation and real environment
## -------------------------------------------------------------------------------------------------

"""
Ver. 2.1.0 (2022-06-11)

This module provides an environment with multivariate state and action spaces 
based on the Gym-based environment 'UR5RandomTargetTask-v0'. 
"""

import sys
import platform
import subprocess
import time
import os
import mlpro
from mlpro.bf.various import Log
from mlpro.wrappers.openai_gym import *
from mlpro.rl.models import *
import numpy as np



## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class UR5JointControl(Environment):
    """
    This environment multivariate space and action spaces by duplicating the
    Gym-based environment 'UR5RandomTargetTask-v0'. 
    """

    C_NAME = 'UR5JointControl'
    C_LATENCY = timedelta(0, 0, 0)
    C_INFINITY = np.finfo(np.float32).max

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_seed=0, p_build=True, p_real=False, p_robot_ip="172.17.0.3", p_reverse_ip="172.17.0.2", 
        p_reverse_port=50001, p_visualize=False, p_logging=True):
        """
        Parameters:
            p_logging       Boolean switch for logging
        """

        # Use ROS as the simulation and real
        if not p_real:
            super().__init__(p_mode=Mode.C_MODE_SIM, p_logging=p_logging)
        else:
            super().__init__(p_mode=Mode.C_MODE_REAL, p_logging=p_logging)
            self._real_ros_state = None

        if p_build:
            self._build()

        try:
            import rospy
            from openai_ros.openai_ros_common import StartOpenAI_ROS_Environment
            from openai_ros.task_envs.task_commons import LoadYamlFileParamsTest
        except (ModuleNotFoundError, ImportError) as e:
            self.log(Log.C_LOG_TYPE_E, e)
            self.log(Log.C_LOG_TYPE_E, "ROS or UR5 Workspace is not installed")
            self.log(Log.C_LOG_TYPE_W, "Please use the guideline here:")
            self.log(Log.C_LOG_TYPE_W, "https://mlpro.readthedocs.io/en/latest/content/rl/env/pool/ur5jointcontrol.html")
        else:
            roscore = subprocess.Popen('roscore')
            rospy.init_node('ur5_lab_training_start', anonymous=True, log_level=rospy.WARN)

            LoadYamlFileParamsTest(rospackage_name="ur5_lab",
                                rel_path_from_package_to_file="config",
                                yaml_file_name="ur5_lab_task_param.yaml")

            ros_ws_path = mlpro.rl.pool.envs.ur5jointcontrol.__file__.replace("/__init__.py", "")
            rospy.set_param('ros_ws_path', ros_ws_path)
            rospy.set_param('sim', not p_real)

            # Init OpenAI_ROS ENV
            if not p_real:
                environment = rospy.get_param('/ur5_lab/simulation_environment')
                rospy.set_param('visualize', p_visualize)
            else:
                environment = rospy.get_param('/ur5_lab/real_environment')
                rospy.set_param('robot_ip', p_robot_ip)
                rospy.set_param('reverse_ip', p_reverse_ip)
                rospy.set_param('reverse_port', p_reverse_port)

            max_step_episode = rospy.get_param('/ur5_lab/max_iterations')

            self._gym_env = StartOpenAI_ROS_Environment(environment, max_step_episode)
            self._gym_env.seed(p_seed)

            self.C_NAME = 'Env "' + self._gym_env.spec.id + '"'

            self._state_space = WrEnvGYM2MLPro.recognize_space(self._gym_env.observation_space)
            self._action_space = WrEnvGYM2MLPro.recognize_space(self._gym_env.action_space)
        

    ## -------------------------------------------------------------------------------------------------
    @staticmethod
    def setup_spaces():
        return None, None


    ## -------------------------------------------------------------------------------------------------
    def _export_action(self, p_action: Action) -> bool:
        try:
            self._real_ros_state = self.simulate_reaction(None, p_action)
        except Exception as e:
            self.log(Log.C_LOG_TYPE_E, e)
            return False
        else:
            return True

    ## -------------------------------------------------------------------------------------------------
    def _import_state(self) -> bool:
        try:
            self._set_state(self._real_ros_state)
        except Exception as e:
            self.log(Log.C_LOG_TYPE_E, e)
            return False
        else:
            return True

    ## -------------------------------------------------------------------------------------------------
    def _reset(self, p_seed=None):
        # 1 Reset Gym environment and determine initial state
        try:
            observation = self._gym_env.reset(seed=p_seed)
        except:
           self._gym_env.seed(p_seed)
           observation = self._gym_env.reset() 
        obs = DataObject(observation)

        # 2 Create state object from Gym observation
        state = State(self._state_space)
        state.set_values(obs.get_data())
        self._set_state(state)

    ## -------------------------------------------------------------------------------------------------
    def _simulate_reaction(self, p_state: State, p_action: Action) -> State:

        # 1 Convert action to Gym syntax
        action_sorted = p_action.get_sorted_values()
        dtype = self._gym_env.action_space.dtype

        if (dtype == np.int32) or (dtype == np.int64):
            action_sorted = action_sorted.round(0)

        if action_sorted.size == 1:
            action_gym = action_sorted.astype(self._gym_env.action_space.dtype)[0]
        else:
            action_gym = action_sorted.astype(self._gym_env.action_space.dtype)

        # 2 Process step of Gym environment
        try:
            observation, reward_gym, done, info = self._gym_env.step(action_gym)
        except:
            observation, reward_gym, done, info = self._gym_env.step(np.atleast_1d(action_gym))

        obs = DataObject(observation)

        # 3 Create state object from Gym observation
        state = State(self._state_space, p_terminal=done)
        state.set_values(obs.get_data())

        # 4 Create reward object
        self._last_reward = Reward(Reward.C_TYPE_OVERALL)
        self._last_reward.set_overall_reward(reward_gym)

        # 5 Return next state
        return state

    ## -------------------------------------------------------------------------------------------------
    def _compute_reward(self, p_state_old: State = None, p_state_new: State = None) -> Reward:
        return self._last_reward

    ## -------------------------------------------------------------------------------------------------
    def _compute_broken(self, p_state: State) -> bool:
        return self.get_broken()

    ## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        pass

    ## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        pass

    ## -------------------------------------------------------------------------------------------------
    def get_cycle_limit(self):
        return self._gym_env._max_episode_steps

    ## -------------------------------------------------------------------------------------------------
    def _compute_success(self, p_state: State) -> bool:
        obs = p_state.get_values()
        close = np.allclose(a=obs[:3],
                            b=obs[3:],
                            atol=0.1)

        if close:
            self._state.set_terminal(True)

        return close

    ## -------------------------------------------------------------------------------------------------
    def _build(self):
        # Check OS
        self.log(Log.C_LOG_TYPE_I, "Checking Operating System ......")
        if platform.system() != "Linux":
            self.log(Log.C_LOG_TYPE_E, "Operating System is not supported!")
            self.log(Log.C_LOG_TYPE_E, "Please use Linux")
            self.log(Log.C_LOG_TYPE_E, "Exiting....")
            sys.exit()
        else:
            self.log(Log.C_LOG_TYPE_S, "Operating System is supported")

        # Check if ROS is installed
        process = subprocess.run("which roscore", shell=True, stdout=subprocess.PIPE)
        output = process.stdout
        self.log(Log.C_LOG_TYPE_I, "Checking ROS Installation ......")
        if output==bytes():
            self.log(Log.C_LOG_TYPE_E, "ROS is not installed!")
            self.log(Log.C_LOG_TYPE_E, "Please install ROS")
            self.log(Log.C_LOG_TYPE_E, "Exiting....")
            sys.exit()
        else:
            self.log(Log.C_LOG_TYPE_S, "ROS is installed")

        import rospkg

        # Check if UR5 Workspace is installed
        installed = False
        rospack = rospkg.RosPack()
        try:
            rospack.get_path("ur5_lab")
        except rospkg.common.ResourceNotFound:
            self.log(Log.C_LOG_TYPE_E, "UR5 Workspace is not installed!")
            self.log(Log.C_LOG_TYPE_W, "If you have ran this script, please CTRL+C and restart terminal")
        else:
            installed = True

        if not installed:
            self.log(Log.C_LOG_TYPE_W, "Building ROS Workspace in 10 Seconds")
            for sec in range(10):
                time.sleep(1)
                self.log(Log.C_LOG_TYPE_W, str(9-sec)+"...")

            ros_workspace = os.path.dirname(mlpro.__file__)+"/rl/pool/envs/ur5jointcontrol"

            # Check dependencies
            command = "cd " + ros_workspace + " && sudo apt-get update && rosdep update && rosdep install --from-paths src --ignore-src -r -y"
            try:
                process = subprocess.check_output(command, shell=True)
            except subprocess.CalledProcessError as e:
                self.log(Log.C_LOG_TYPE_E, "Installing Dependencies Failed")
                sys.exit()

            # Build Workspace
            command = "cd " + ros_workspace + " && catkin_make"
            try:
                process = subprocess.check_output(command, shell=True)
            except subprocess.CalledProcessError as e:
                self.log(Log.C_LOG_TYPE_E, "Build Failed")
                sys.exit()

            self.log(Log.C_LOG_TYPE_S, "Successfully Built")
            command = "echo 'source "+ros_workspace+"/devel/setup.bash"+"' >> ~/.bashrc"
            process = subprocess.run(command, shell=True)
            self.log(Log.C_LOG_TYPE_W, "Please restart your terminal and run the Howto script again")
            sys.exit()
        else:
            self.log(Log.C_LOG_TYPE_S, "UR5 Workspace is installed")


## -----------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro
## -- Module  : BGLP_GT
## -----------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-06-07  0.0.0     SY       Creation
## -- 2021-08-31  1.0.0     SY       Release of first version
## -- 2021-09-01  1.0.1     SY       Minor improvements, code cleaning, add descriptions
## -- 2021-09-06  1.0.2     SY       Minor improvements
## -- 2021-09-11  1.0.3     MRD      Change Header information to match our new library name
## -- 2021-11-29  1.0.4     SY       Enable batch production scenario
## -----------------------------------------------------------------------------

"""
Ver. 1.0.4 (2021-11-29)

This module provides an environment of Bulk Good Laboratory Plant (BGLP)
following GT interface. This module provides game board classed based on BGLP
environment of the reinforcement learning pool.
"""

from mlpro.rl.models import Reward
from mlpro.rl.pool.envs.bglp import BGLP
from mlpro.gt.models import *

        
        
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------

class BGLP_GT(BGLP, GameBoard):
    """
    Game theoretical pendant for the reinforcement learning environment class BGLP.
    """

    C_NAME          = 'BGLP_GT'

    def __init__(self, p_logging=True,t_step=0.5, t_set=10.0, demand=0.1,
                 lr_margin=1.0, lr_demand=4.0, lr_energy=0.0010, margin_p=[0.2,0.8,4],
                 prod_target=10000, prod_scenario='continuous'):
        BGLP.__init__(self, p_reward_type=Reward.C_TYPE_EVERY_AGENT, p_logging=p_logging,
                      t_step=t_step, t_set=t_set, demand=demand, lr_margin=lr_margin,
                      lr_demand=lr_demand, lr_energy=lr_energy, margin_p=margin_p,
                      prod_target=prod_target, prod_scenario=prod_scenario)





#!/usr/bin/env python
import sys
import time
import unittest

import rospy
import actionlib
import std_msgs.msg
from control_msgs.msg import (
    FollowJointTrajectoryAction,
    FollowJointTrajectoryGoal,
    FollowJointTrajectoryResult,
    JointTolerance)
from ur_dashboard_msgs.msg import SetModeAction, SetModeGoal, RobotMode
from std_srvs.srv import Trigger, TriggerRequest
import tf
from trajectory_msgs.msg import JointTrajectoryPoint
from ur_msgs.srv import SetIO, SetIORequest
from ur_msgs.msg import IOStates

from cartesian_control_msgs.msg import (
    FollowCartesianTrajectoryAction,
    FollowCartesianTrajectoryGoal,
    FollowCartesianTrajectoryResult,
    CartesianTrajectoryPoint)
import geometry_msgs.msg

from controller_manager_msgs.srv import SwitchControllerRequest, SwitchController

PKG = 'ur_robot_driver'
NAME = 'integration_test'
ALL_CONTROLLERS = [
        "scaled_pos_joint_traj_controller",
        "pos_joint_traj_controller",
        "scaled_vel_joint_traj_controller",
        "vel_joint_traj_controller",
        "joint_group_vel_controller",
        "forward_joint_traj_controller",
        "forward_cartesian_traj_controller",
        "twist_controller",
        "pose_based_cartesian_traj_controller",
        "joint_based_cartesian_traj_controller",
        ]


class IntegrationTest(unittest.TestCase):
    def __init__(self, *args):
        super(IntegrationTest, self).__init__(*args)

        self.init_robot()

    def init_robot(self):
        """Make sure the robot is booted and ready to receive commands"""

        rospy.init_node('ur_robot_driver_integration_test')
        timeout = rospy.Duration(30)

        self.set_mode_client = actionlib.SimpleActionClient(
            '/ur_hardware_interface/set_mode', SetModeAction)
        if not self.set_mode_client.wait_for_server(timeout):
            self.fail(
                "Could not reach set_mode action. Make sure that the driver is actually running."
                " Msg: {}".format(err))

        self.trajectory_client = actionlib.SimpleActionClient(
            'follow_joint_trajectory', FollowJointTrajectoryAction)
        if not self.trajectory_client.wait_for_server(timeout):
            self.fail(
                "Could not reach controller action. Make sure that the driver is actually running."
                " Msg: {}".format(err))

        self.cartesian_passthrough_trajectory_client = actionlib.SimpleActionClient(
            'forward_cartesian_trajectory', FollowCartesianTrajectoryAction)
        if not self.cartesian_passthrough_trajectory_client.wait_for_server(timeout):
            self.fail(
                "Could not reach cartesian passthrough controller action. Make sure that the driver is actually running."
                " Msg: {}".format(err))

        self.joint_passthrough_trajectory_client = actionlib.SimpleActionClient(
            'forward_joint_trajectory', FollowJointTrajectoryAction)
        if not self.joint_passthrough_trajectory_client.wait_for_server(timeout):
            self.fail(
                "Could not reach joint passthrough controller action. Make sure that the driver is actually running."
                " Msg: {}".format(err))

        self.cartesian_trajectory_client = actionlib.SimpleActionClient(
            'follow_cartesian_trajectory', FollowCartesianTrajectoryAction)
        if not self.cartesian_trajectory_client.wait_for_server(timeout):
            self.fail(
                "Could not reach cartesian controller action. Make sure that the driver is actually running."
                " Msg: {}".format(err))

        self.set_io_client = rospy.ServiceProxy('/ur_hardware_interface/set_io', SetIO)
        try:
            self.set_io_client.wait_for_service(timeout)
        except rospy.exceptions.ROSException as err:
            self.fail(
                "Could not reach SetIO service. Make sure that the driver is actually running."
                " Msg: {}".format(err))

        self.switch_controllers_client = rospy.ServiceProxy('/controller_manager/switch_controller',
                SwitchController)
        try:
            self.switch_controllers_client.wait_for_service(timeout)
        except rospy.exceptions.ROSException as err:
            self.fail(
                "Could not reach controller switch service. Make sure that the driver is actually running."
                " Msg: {}".format(err))

        self.send_program_srv = rospy.ServiceProxy('/ur_hardware_interface/resend_robot_program',
                                                   Trigger)
        try:
            self.send_program_srv.wait_for_service(timeout)
        except rospy.exceptions.ROSException as err:
            self.fail(
                "Could not reach resend_robot_program service. Make sure that the driver is "
                "actually running in headless mode."
                " Msg: {}".format(err))

        self.script_publisher = rospy.Publisher("/ur_hardware_interface/script_command", std_msgs.msg.String, queue_size=1)
        self.tf_listener = tf.TransformListener()
        self.twist_pub = rospy.Publisher("/twist_controller/command", geometry_msgs.msg.Twist, queue_size=1)

    def set_robot_to_mode(self, target_mode):
        goal = SetModeGoal()
        goal.target_robot_mode = target_mode
        goal.play_program = False # we use headless mode during tests
        # This might be a bug to hunt down. We have to reset the program before calling `resend_robot_program`
        goal.stop_program = True

        self.set_mode_client.send_goal(goal)
        self.set_mode_client.wait_for_result()
        return self.set_mode_client.get_result().success


    def test_joint_trajectory_position_interface(self):
        """Test robot movement"""
        #### Power cycle the robot in order to make sure it is running correctly####
        self.assertTrue(self.set_robot_to_mode(RobotMode.POWER_OFF))
        rospy.sleep(0.5)
        self.assertTrue(self.set_robot_to_mode(RobotMode.RUNNING))
        rospy.sleep(0.5)

        self.send_program_srv.call()
        rospy.sleep(0.5) # TODO properly wait until the controller is running
        self.switch_on_controller("scaled_pos_joint_traj_controller")

        goal = FollowJointTrajectoryGoal()

        goal.trajectory.joint_names = ["elbow_joint", "shoulder_lift_joint", "shoulder_pan_joint",
                                       "wrist_1_joint", "wrist_2_joint", "wrist_3_joint"]
        position_list = [[0.0 for i in range(6)]]
        position_list.append([-0.5 for i in range(6)])
        position_list.append([-1.0 for i in range(6)])
        duration_list = [6.0, 9.0, 12.0]

        for i, position in enumerate(position_list):
            point = JointTrajectoryPoint()
            point.positions = position
            point.time_from_start = rospy.Duration(duration_list[i])
            goal.trajectory.points.append(point)

        rospy.loginfo("Sending simple goal")

        self.trajectory_client.send_goal(goal)
        self.trajectory_client.wait_for_result()

        self.assertEqual(self.trajectory_client.get_result().error_code,
                         FollowJointTrajectoryResult.SUCCESSFUL)
        rospy.loginfo("Received result SUCCESSFUL")

        """Test trajectory server. This is more of a validation test that the testing suite does the
        right thing."""
        goal = FollowJointTrajectoryGoal()

        goal.trajectory.joint_names = ["elbow_joint", "shoulder_lift_joint", "shoulder_pan_joint",
                                       "wrist_1_joint", "wrist_2_joint", "wrist_3_joint"]
        position_list = [[0.0 for i in range(6)]]
        position_list.append([-0.5 for i in range(6)])
        # Create illegal goal by making the second point come earlier than the first
        duration_list = [6.0, 3.0]

        for i, position in enumerate(position_list):
            point = JointTrajectoryPoint()
            point.positions = position
            point.time_from_start = rospy.Duration(duration_list[i])
            goal.trajectory.points.append(point)

        rospy.loginfo("Sending illegal goal")
        self.trajectory_client.send_goal(goal)
        self.trajectory_client.wait_for_result()

        # As timings are illegal, we expect the result to be INVALID_GOAL
        self.assertEqual(self.trajectory_client.get_result().error_code,
                         FollowJointTrajectoryResult.INVALID_GOAL)
        rospy.loginfo("Received result INVALID_GOAL")

        """Test robot movement"""
        goal = FollowJointTrajectoryGoal()

        goal.trajectory.joint_names = ["elbow_joint", "shoulder_lift_joint", "shoulder_pan_joint",
                                       "wrist_1_joint", "wrist_2_joint", "wrist_3_joint"]
        position_list = [[0.0 for i in range(6)]]
        position_list.append([-1.0 for i in range(6)])
        duration_list = [6.0, 6.5]

        for i, position in enumerate(position_list):
            point = JointTrajectoryPoint()
            point.positions = position
            point.time_from_start = rospy.Duration(duration_list[i])
            goal.trajectory.points.append(point)

        rospy.loginfo("Sending scaled goal without time restrictions")
        self.trajectory_client.send_goal(goal)
        self.trajectory_client.wait_for_result()

        self.assertEqual(self.trajectory_client.get_result().error_code,
                         FollowJointTrajectoryResult.SUCCESSFUL)
        rospy.loginfo("Received result SUCCESSFUL")

        # Now do the same again, but with a goal time constraint
        rospy.loginfo("Sending scaled goal with time restrictions")
        goal.goal_time_tolerance = rospy.Duration(0.01)
        self.trajectory_client.send_goal(goal)
        self.trajectory_client.wait_for_result()

        self.assertEqual(self.trajectory_client.get_result().error_code,
                         FollowJointTrajectoryResult.GOAL_TOLERANCE_VIOLATED)
        rospy.loginfo("Received result GOAL_TOLERANCE_VIOLATED")


    def test_set_io(self):
        """Test to set an IO and check whether it has been set."""
        maximum_messages = 5
        pin = 0
        self.assertEqual(maximum_messages, 5)

        self.set_io_client(1, pin, 0)
        messages = 0
        pin_state = True

        while(pin_state):
            if messages >= maximum_messages:
                self.fail("Could not read desired state after {} messages.".format(maximum_messages))
            io_state = rospy.wait_for_message('/ur_hardware_interface/io_states', IOStates)
            pin_state = io_state.digital_out_states[pin].state
            messages += 1
        self.assertEqual(pin_state, 0)

        self.set_io_client(SetIORequest.FUN_SET_DIGITAL_OUT, pin, 1)
        messages = 0
        pin_state = False

        while(not pin_state):
            if messages >= maximum_messages:
                self.fail("Could not read desired state after {} messages.".format(maximum_messages))
            io_state = rospy.wait_for_message('/ur_hardware_interface/io_states', IOStates)
            pin_state = io_state.digital_out_states[pin].state
            messages += 1
        self.assertEqual(pin_state, 1)

    def test_cartesian_passthrough(self):
        #### Power cycle the robot in order to make sure it is running correctly####
        self.assertTrue(self.set_robot_to_mode(RobotMode.POWER_OFF))
        rospy.sleep(0.5)
        self.assertTrue(self.set_robot_to_mode(RobotMode.RUNNING))
        rospy.sleep(0.5)

        # Make sure the robot is at a valid start position for our cartesian motions
        self.script_publisher.publish("movej([1, -1.7, -1.7, -1, -1.57, -2])")
        # As we don't have any feedback from that interface, sleep for a while
        rospy.sleep(5)


        self.send_program_srv.call()
        rospy.sleep(0.5) # TODO properly wait until the controller is running

        self.switch_on_controller("forward_cartesian_traj_controller")

        position_list = [geometry_msgs.msg.Vector3(0.4,0.4,0.4)]
        position_list.append(geometry_msgs.msg.Vector3(0.5,0.5,0.5))
        duration_list = [3.0, 6.0]
        goal = FollowCartesianTrajectoryGoal()

        for i, position in enumerate(position_list):
            point = CartesianTrajectoryPoint()
            point.pose = geometry_msgs.msg.Pose(position, geometry_msgs.msg.Quaternion(0,0,0,1))
            point.time_from_start = rospy.Duration(duration_list[i])
            goal.trajectory.points.append(point)
        self.cartesian_passthrough_trajectory_client.send_goal(goal)
        self.cartesian_passthrough_trajectory_client.wait_for_result()
        self.assertEqual(self.cartesian_passthrough_trajectory_client.get_result().error_code,
                         FollowCartesianTrajectoryResult.SUCCESSFUL)
        rospy.loginfo("Received result SUCCESSFUL")

    def test_joint_passthrough(self):
        #### Power cycle the robot in order to make sure it is running correctly####
        self.assertTrue(self.set_robot_to_mode(RobotMode.POWER_OFF))
        rospy.sleep(0.5)
        self.assertTrue(self.set_robot_to_mode(RobotMode.RUNNING))
        rospy.sleep(0.5)

        self.send_program_srv.call()
        rospy.sleep(0.5) # TODO properly wait until the controller is running

        self.switch_on_controller("forward_joint_traj_controller")
        goal = FollowJointTrajectoryGoal()

        goal.trajectory.joint_names = ["elbow_joint", "shoulder_lift_joint", "shoulder_pan_joint",
                                       "wrist_1_joint", "wrist_2_joint", "wrist_3_joint"]

        position_list = [[0,-1.57,-1.57,0,0,0]]
        position_list.append([0.2,-1.57,-1.57,0,0,0])
        position_list.append([-0.5,-1.57,-1.2,0,0,0])
        duration_list = [3.0, 7.0, 10.0]

        for i, position in enumerate(position_list):
            point = JointTrajectoryPoint()
            point.positions = position
            point.time_from_start = rospy.Duration(duration_list[i])
            goal.trajectory.points.append(point)
        for i, joint_name in enumerate(goal.trajectory.joint_names):
            goal.goal_tolerance.append(JointTolerance(joint_name, 0.2, 0.2, 0.2))

        goal.goal_time_tolerance = rospy.Duration(0.6)
        self.joint_passthrough_trajectory_client.send_goal(goal)
        self.joint_passthrough_trajectory_client.wait_for_result()

        self.assertEqual(self.joint_passthrough_trajectory_client.get_result().error_code,
                         FollowJointTrajectoryResult.SUCCESSFUL)
        rospy.loginfo("Received result SUCCESSFUL")

    def test_cartesian_trajectory_pose_interface(self):
        #### Power cycle the robot in order to make sure it is running correctly####
        self.assertTrue(self.set_robot_to_mode(RobotMode.POWER_OFF))
        rospy.sleep(0.5)
        self.assertTrue(self.set_robot_to_mode(RobotMode.RUNNING))
        rospy.sleep(0.5)

        # Make sure the robot is at a valid start position for our cartesian motions
        self.script_publisher.publish("movej([1, -1.7, -1.7, -1, -1.57, -2])")
        # As we don't have any feedback from that interface, sleep for a while
        rospy.sleep(5)


        self.send_program_srv.call()
        rospy.sleep(0.5) # TODO properly wait until the controller is running

        self.switch_on_controller("pose_based_cartesian_traj_controller")

        position_list = [geometry_msgs.msg.Vector3(0.4,0.4,0.4)]
        position_list.append(geometry_msgs.msg.Vector3(0.5,0.5,0.5))
        duration_list = [3.0, 6.0]
        goal = FollowCartesianTrajectoryGoal()

        for i, position in enumerate(position_list):
            point = CartesianTrajectoryPoint()
            point.pose = geometry_msgs.msg.Pose(position, geometry_msgs.msg.Quaternion(0,0,0,1))
            point.time_from_start = rospy.Duration(duration_list[i])
            goal.trajectory.points.append(point)
        self.cartesian_trajectory_client.send_goal(goal)
        self.cartesian_trajectory_client.wait_for_result()
        self.assertEqual(self.cartesian_trajectory_client.get_result().error_code,
                         FollowCartesianTrajectoryResult.SUCCESSFUL)
        rospy.loginfo("Received result SUCCESSFUL")

    def test_twist_interface(self):
        #### Power cycle the robot in order to make sure it is running correctly####
        self.assertTrue(self.set_robot_to_mode(RobotMode.POWER_OFF))
        rospy.sleep(0.5)
        self.assertTrue(self.set_robot_to_mode(RobotMode.RUNNING))
        rospy.sleep(0.5)

        # Make sure the robot is at a valid start position for our cartesian motions
        self.script_publisher.publish("movej([1, -1.7, -1.7, -1, -1.57, -2])")
        # As we don't have any feedback from that interface, sleep for a while
        rospy.sleep(5)


        self.send_program_srv.call()
        rospy.sleep(0.5) # TODO properly wait until the controller is running

        self.switch_on_controller("twist_controller")

        # Lookup tcp in base_frame
        (trans_start, rot_start) = self.tf_listener.lookupTransform('base', 'tool0_controller', rospy.Time(0))

        twist = geometry_msgs.msg.Twist()
        twist.linear.x = 0.1
        twist.linear.y = 0.0
        twist.linear.z = 0.0
        twist.angular.x = 0.0
        twist.angular.y = 0.0
        twist.angular.z = 0.0

        # publish twist
        self.twist_pub.publish(twist)

        # wait 1 sec
        rospy.sleep(1)

        # stop robot
        twist.linear.x = 0.0
        self.twist_pub.publish(twist)

        (trans_end, rot_end) = self.tf_listener.lookupTransform('base', 'tool0_controller', rospy.Time(0))

        self.assertAlmostEqual(rot_start[0], rot_end[0], delta=3e-6)
        self.assertAlmostEqual(rot_start[1], rot_end[1], delta=1e-6)
        self.assertAlmostEqual(rot_start[2], rot_end[2], delta=1e-6)
        self.assertAlmostEqual(trans_start[1], trans_end[1], delta=1e-6)
        self.assertAlmostEqual(trans_start[2], trans_end[2], delta=1e-6)
        self.assertTrue(trans_end[0] > trans_start[0])


    def switch_on_controller(self, controller_name):
        """Switches on the given controller stopping all other known controllers with best_effort
        strategy."""
        srv = SwitchControllerRequest()
        srv.stop_controllers = ALL_CONTROLLERS
        srv.start_controllers = [controller_name]
        srv.strictness = SwitchControllerRequest.BEST_EFFORT
        result = self.switch_controllers_client(srv)
        self.assertTrue(result.ok)

if __name__ == '__main__':
    import rostest
    rostest.run(PKG, NAME, IntegrationTest, sys.argv)



## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro
## -- Module  : various
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-04-16  0.0.0     DA       Creation
## -- 2021-05-29  1.0.0     DA       Release of first version
## -- 2021-06-16  1.1.0     SY       Adding the first version of data storing,
## --                                data plotting, and data saving classes
## -- 2021-06-17  1.2.0     DA       New abstract classes Loadable, Saveable
## -- 2021-06-21  1.3.0     SY       Add extensions in classes Loadable,
## --                                Saveable, DataPlotting & DataStoring.
## -- 2021-07-01  1.4.0     SY       Extend save/load functionalities
## -- 2021-08-20  1.5.0     DA       Added property class Plottable
## -- 2021-08-28  1.5.1     DA       Added constant C_VAR0 to class DataStoring
## -- 2021-09-11  1.5.0     MRD      Change Header information to match our new library name
## -- 2021-10-06  1.5.2     DA       Moved class DataStoring to new module mlpro.bf.data.py and
## --                                classes DataPlotting, Plottable to new module mlpro.bf.plot.py
## -- 2021-10-07  1.6.0     DA       Class Log: 
## --                                - colored text depending on log type 
## --                                - new method set_log_level()
## -- 2021-10-25  1.7.0     SY       Add new class ScientificObject
## -- 2021-11-03  1.7.1     DA       Class Log: new type C_LOG_TYPE_SUCCESS for success messages 
## -- 2021-11-15  1.7.2     DA       Class Log: 
## --                                - method set_log_level() removed
## --                                - parameter p_logging is the new log level now
## --                                Class Saveable: new constant C_SUFFIX
## -- 2021-12-07  1.7.3     SY       Add a new attribute in ScientificObject
## -- 2021-12-31  1.7.4     DA       Class Log: udpated docstrings
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.7.4 (2021-12-31)

This module provides various classes with elementry functionalities for reuse in higher level classes. 
For example: logging, load/save, timer, ...
"""

from datetime import datetime, timedelta
from time import sleep
import dill as pkl
import os
from mlpro.bf.exceptions import *


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Loadable:
    """
    This abstract class adds the ability to be loadable to inherited classes. 
    """

    ## -------------------------------------------------------------------------------------------------
    @staticmethod
    def load(p_path, p_filename):
        """
        Loads content from the given path and file name. If file does not exist, it returns None.

        Parameters:
            p_path          Path that contains the file 
            p_filename      File name

        Returns: 
            A loaded object, if file content was loaded successfully. None otherwise.
        """

        if not os.path.exists(p_path + os.sep + p_filename):
            return None

        return pkl.load(open(p_path + os.sep + p_filename, 'rb'))


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Saveable:
    """
    This abstract class adds the ability to be saveable to inherited classes. The filename can be
    generated internally by implementing the method get_filename() or provided from outside otherwise.
    """

    C_SUFFIX = '.pkl'

    ## -------------------------------------------------------------------------------------------------
    def generate_filename(self):
        """
        To be redefined in case of use of internal generated file names.

        Returns: 
            Returns an internal unique filename. 
        """

        raise NotImplementedError

    ## -------------------------------------------------------------------------------------------------
    def save(self, p_path, p_filename=None) -> bool:
        """
        Saves content to the given path and file name. If file name is None, a unique file name will
        be generated by calling method generate_filename(). If it returns False then the saving method is failed. 

        Parameters:
            p_path          Path where file will be saved
            p_filename      File name (if None an internal filename will be generated)

        Returns: 
            True, if file content was saved successfully. False otherwise.
        """

        if (p_filename is not None) and (p_filename != ''):
            self.filename = p_filename
        else:
            self.filename = self.generate_filename()

        if self.filename is None:
            return False

        try:
            if not os.path.exists(p_path):
                os.makedirs(p_path)
            pkl.dump(self, open(p_path + os.sep + self.filename, "wb"))
            return True
        except:
            return False


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class LoadSave(Loadable, Saveable):
    """
    This abstract class adds the ability to be loadable and saveable to inherited classes. The 
    filename can be generated internally by implementing the method generate_filename() or provided 
    from outside otherwise. See classes Loadable and Saveable for further information.
    """

    pass


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Log:
    """
    This class adds elementry log functionality to inherited classes.
    """

    C_TYPE = '????'
    C_NAME = '????'

    # Types of log lines
    C_LOG_TYPE_I = 'I'  # Information
    C_LOG_TYPE_W = 'W'  # Warning
    C_LOG_TYPE_E = 'E'  # Error
    C_LOG_TYPE_S = 'S'  # Success / Milestone

    C_LOG_TYPES = [C_LOG_TYPE_I, C_LOG_TYPE_W, C_LOG_TYPE_E, C_LOG_TYPE_S]

    C_COL_WARNING = '\033[93m'  # Yellow
    C_COL_ERROR = '\033[91m'  # Red
    C_COL_SUCCESS = '\033[32m'  # Green
    C_COL_RESET = '\033[0m'  # Reset color

    # Log levels
    C_LOG_ALL = True
    C_LOG_NOTHING = False
    C_LOG_WE = C_LOG_TYPE_W
    C_LOG_E = C_LOG_TYPE_E

    C_LOG_LEVELS = [C_LOG_ALL, C_LOG_NOTHING, C_LOG_WE, C_LOG_E]

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_logging=C_LOG_ALL):
        """
        Parameters:
            p_logging     Log level (see constants for log levels)
        """

        self.switch_logging(p_logging)
        self.log(self.C_LOG_TYPE_I, 'Instantiated')

    ## -------------------------------------------------------------------------------------------------
    def switch_logging(self, p_logging):
        """
        Sets new log level. 

        Parameters
        ----------
        p_logging     
            Log level (constant C_LOG_LEVELS contains valid values)

        """

        if p_logging not in self.C_LOG_LEVELS: raise ParamError('Wrong log level. See class Log for valid log levels')
        self._level = p_logging

    # ## -------------------------------------------------------------------------------------------------
    #     def set_log_level(self, p_level):
    #         """
    #         Sets the log level.

    #         Parameters:
    #             p_level         Possible values are
    #                             C_LOG_TYPE_I (everything will be looged)
    #                             C_LOG_TYPE_W (warnings and errors will be logged)
    #                             C_LOG_TYPE_E (only errors will be logged)
    #         """

    #         if p_level in self.C_LOG_TYPES:
    #             self._level = p_level
    #         else:
    #             raise ParamError('Wrong log level. Please use constants C_LOG_TYPE_* of class Log')

    ## -------------------------------------------------------------------------------------------------
    def log(self, p_type, *p_args):
        """
        Writes log line to standard output in format:
        yyyy-mm-dd  hh:mm:ss.mmmmmm  [p_type  C_TYPE C_NAME]: [p_args] 

        Parameters:
            p_type      type of log entry
            p_args      log informations

        Returns: 
            Nothing
        """

        if not self._level: return

        if self._level == self.C_LOG_WE:
            if (p_type == self.C_LOG_TYPE_I) or (p_type == self.C_LOG_TYPE_S): return
        elif self._level == self.C_LOG_E:
            if (p_type == self.C_LOG_TYPE_I) or (p_type == self.C_LOG_TYPE_S) or (p_type == self.C_LOG_TYPE_W): return

        now = datetime.now()

        if p_type == self.C_LOG_TYPE_W:
            col = self.C_COL_WARNING
        elif p_type == self.C_LOG_TYPE_E:
            col = self.C_COL_ERROR
        elif p_type == self.C_LOG_TYPE_S:
            col = self.C_COL_SUCCESS
        else:
            col = self.C_COL_RESET

        print(col + '%04d-%02d-%02d  %02d:%02d:%02d.%06d ' % (
        now.year, now.month, now.day, now.hour, now.minute, now.second, now.microsecond),
              p_type + '  ' + self.C_TYPE + ' ' + self.C_NAME + ':', *p_args, self.C_COL_RESET)


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Timer:
    """
    Timer class in two time modes (real/virtual) and with simple lap management.
    """

    C_MODE_REAL = 0  # Real time
    C_MODE_VIRTUAL = 1  # Virtual time

    C_LAP_LIMIT = 999999

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_mode, p_lap_duration: timedelta, p_lap_limit=C_LAP_LIMIT) -> None:
        """
        Parameters:
            p_mode          C_MODE_REAL for real time mode,
                            C_MODE_VIRTUAL for virtual time mode
            p_lap_duration  Duration of a single lap
            p_lap_limit     maximum number of laps until the lap counter restarts with 0  
        """

        self.mode = p_mode
        self.lap_duration = p_lap_duration

        if p_lap_limit == 0:
            self.lap_limit = self.C_LAP_LIMIT
        else:
            self.lap_limit = p_lap_limit

        self.reset()

    ## -------------------------------------------------------------------------------------------------
    def reset(self) -> None:
        """
        Resets timer.

        Returns: 
            Nothing
        """

        self.time = timedelta(0, 0, 0)
        self.lap_time = timedelta(0, 0, 0)
        self.lap_id = 0

        if self.mode == self.C_MODE_REAL:
            self.timer_start_real = datetime.now()
            self.lap_start_real = self.timer_start_real
            self.time_real = self.timer_start_real

    ## -------------------------------------------------------------------------------------------------
    def get_time(self) -> timedelta:
        if self.mode == self.C_MODE_REAL:
            self.time_real = datetime.now()
            self.time = self.time_real - self.timer_start_real

        return self.time

    ## -------------------------------------------------------------------------------------------------
    def get_lap_time(self) -> timedelta:
        if self.mode == self.C_MODE_REAL:
            self.lap_time = datetime.now() - self.lap_start_real

        return self.lap_time

    ## -------------------------------------------------------------------------------------------------
    def get_lap_id(self):
        return self.lap_id

    ## -------------------------------------------------------------------------------------------------
    def add_time(self, p_delta: timedelta):
        if self.mode == self.C_MODE_VIRTUAL:
            self.lap_time = self.lap_time + p_delta
            self.time = self.time + p_delta

    ## -------------------------------------------------------------------------------------------------
    def finish_lap(self) -> bool:
        """
        Finishes the current lap. In timer mode C_MODE_REAL the remaining time
        until the end of the lap will be paused. 

        Returns: 
            True, if the remaining time to the next lap was positive. False, if 
            the timer timed out.
        """

        timeout = False

        # Compute delay until next lap
        delay = self.lap_duration - self.get_lap_time()

        # Check for timeout
        if delay < timedelta(0, 0, 0):
            timeout = True
            delay = timedelta(0, 0, 0)

        # Handle delay depending on timer mode
        if self.mode == self.C_MODE_REAL:
            # Wait until next lap start
            sleep(delay.total_seconds())
        else:
            # Just set next lap start time
            self.time = self.time + delay

        # Update lap data
        self.lap_id = divmod(self.lap_id + 1, self.lap_limit)[1]
        self.lap_time = timedelta(0, 0, 0)
        self.lap_start_real = datetime.now()

        return not timeout


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class TStamp:
    """
    This class provides elementry time stamp functionality for inherited classes.
    """

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_tstamp: timedelta = None):
        self.set_tstamp(p_tstamp)

    ## -------------------------------------------------------------------------------------------------
    def get_tstamp(self) -> timedelta:
        return self.tstamp

    ## -------------------------------------------------------------------------------------------------
    def set_tstamp(self, p_tstamp: timedelta):
        self.tstamp = p_tstamp


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class  ScientificObject:
    """
    This class provides elementary functionality for storing a scientific reference.
    """

    C_SCIREF_TYPE_NONE = None
    C_SCIREF_TYPE_ARTICLE = "Journal Article"
    C_SCIREF_TYPE_BOOK = "Book"
    C_SCIREF_TYPE_ONLINE = "Online"
    C_SCIREF_TYPE_PROCEEDINGS = "Proceedings"
    C_SCIREF_TYPE_TECHREPORT = "Technical Report"
    C_SCIREF_TYPE_UNPUBLISHED = "Unpublished"

    C_SCIREF_TYPE = C_SCIREF_TYPE_NONE
    C_SCIREF_AUTHOR = None
    C_SCIREF_TITLE = None
    C_SCIREF_JOURNAL = None
    C_SCIREF_ABSTRACT = None
    C_SCIREF_VOLUME = None
    C_SCIREF_NUMBER = None
    C_SCIREF_PAGES = None
    C_SCIREF_YEAR = None
    C_SCIREF_MONTH = None
    C_SCIREF_DAY = None
    C_SCIREF_DOI = None
    C_SCIREF_KEYWORDS = None
    C_SCIREF_ISBN = None
    C_SCIREF_SERIES = None
    C_SCIREF_PUBLISHER = None
    C_SCIREF_CITY = None
    C_SCIREF_COUNTRY = None
    C_SCIREF_URL = None
    C_SCIREF_CHAPTER = None
    C_SCIREF_BOOKTITLE = None
    C_SCIREF_INSTITUTION = None
    C_SCIREF_CONFERENCE = None
    C_SCIREF_NOTES = None
    C_SCIREF_EDITOR = None



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_rl_003_train_agent_with_own_policy_on_gym_environment.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-06-03  0.0.0     DA       Creation
## -- 2021-06-06  1.0.0     DA       Released first version
## -- 2021-06-25  1.1.0     DA       Extended by data logging/storing (user home directory)
## -- 2021-07-06  1.1.1     SY       Bugfix due to method Training.save_data() update
## -- 2021-08-28  1.2.0     DA       Introduced Policy
## -- 2021-09-11  1.2.0     MRD      Change Header information to match our new library name
## -- 2021-09-28  1.2.1     SY       Updated due to implementation of method get_cycle_limits()
## -- 2021-09-29  1.2.2     SY       Change name: WrEnvGym to WrEnvGYM2MLPro
## -- 2021-10-06  1.2.3     DA       Refactoring 
## -- 2021-10-18  1.2.4     DA       Refactoring 
## -- 2021-11-15  1.3.0     DA       Refactoring 
## -- 2021-12-03  1.3.1     DA       Refactoring 
## -- 2021-12-07  1.3.2     DA       Refactoring 
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.3.2 (2021-12-07)

This module shows how to train an agent with a custom policy inside on an OpenAI Gym environment using the fhswf_at_ml framework.
"""


from sys import path
from mlpro.bf.math import *
from mlpro.rl.models import *
from mlpro.wrappers.openai_gym import WrEnvGYM2MLPro
import gym
import random
from pathlib import Path




# 1 Implement your own agent policy
class MyPolicy (Policy):

    C_NAME      = 'MyPolicy'

    def set_random_seed(self, p_seed=None):
        random.seed(p_seed)


    def compute_action(self, p_state: State) -> Action:
        # 1 Create a numpy array for your action values 
        my_action_values = np.zeros(self._action_space.get_num_dim())

        # 2 Computing action values is up to you...
        for d in range(self._action_space.get_num_dim()):
            my_action_values[d] = random.random() 

        # 3 Return an action object with your values
        return Action(self._id, self._action_space, my_action_values)


    def _adapt(self, *p_args) -> bool:
        # 1 Adapting the internal policy is up to you...
        self.log(self.C_LOG_TYPE_I, 'Sorry, I am a stupid agent...')

        # 2 Only return True if something has been adapted...
        return False




# 2 Implement your own RL scenario
class MyScenario (RLScenario):

    C_NAME      = 'Matrix'

    def _setup(self, p_mode, p_ada, p_logging):
        # 1 Setup environment
        gym_env     = gym.make('CartPole-v1')
        self._env   = WrEnvGYM2MLPro(gym_env, p_logging=p_logging) 

        # 2 Setup and return standard single-agent with own policy
        return Agent(
                p_policy=MyPolicy(
                    p_observation_space=self._env.get_state_space(),
                    p_action_space=self._env.get_action_space(),
                    p_buffer_size=10,
                    p_ada=p_ada,
                    p_logging=p_logging
                ),    
            p_envmodel=None,
            p_name='Smith',
            p_ada=p_ada,
            p_logging=p_logging
        )




# 3 Create scenario and start training

if __name__ == "__main__":
    # 3.1 Parameters for demo mode
    cycle_limit = 500
    logging     = Log.C_LOG_WE
    visualize   = True
    path        = str(Path.home())
 
else:
    # 3.2 Parameters for internal unit test
    cycle_limit = 50
    logging     = Log.C_LOG_NOTHING
    visualize   = False
    path        = None


# 3.3 Create and run training object
training = RLTraining(
        p_scenario_cls=MyScenario,
        p_cycle_limit=cycle_limit,
        p_path=path,
        p_visualize=visualize,
        p_logging=logging )

training.run()


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.bf.ui.scui.pool
## -- Module  : iis
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-06-20  0.0.0     DA       Creation
## -- 2021-07-04  1.0.0     DA       Released first version
## -- 2021-07-07  1.0.1     DA       Minor corrections related to window size and update rate
## -- 2021-09-11  1.0.1     MRD      Change Header information to match our new library name
## -- 2022-03-21  1.0.2     SY       Refactoring following class Dimensions update 
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.2 (2022-03-21)

This module provides the reusable SciUI component 'Interactive Input Space'. It represents a 2D or 3D 
input space based on a mathematical space and allows manual (2D) and automated (2D,3D) input creation.
It consists of time depending subplots for each input dimension and a 2D/3D main subplot for the input
space. To be reused and enriched in own SciUI scenarios.
"""


from math import sin, cos, pi
from tkinter import *
import tkinter
from tkinter.ttk import Combobox
from mpl_toolkits.mplot3d import Axes3D

from mlpro.bf.ui.sciui.framework import *
from mlpro.bf.ui.sciui.pool.iisbenchmark import *
from mlpro.bf.math import ESpace





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class IISIVar(SciUISubplot2D):
    """
    Subplot for input variable.
    """

    C_NAME                      = 'IIS Input Variable'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        if self.shared_db.iis_ispace.get_num_dim() == 2:
            self.C_AX_RECTANGLE              = [0.10, 0.11, 0.85, 0.85]
        else:
            self.C_AX_RECTANGLE              = [0.10, 0.15, 0.85, 0.80]

        super().init_component()

        self.plot               = None
        self.min_val            = 1
        self.max_val            = 0
        self.val_range_changed  = False


## -------------------------------------------------------------------------------------------------
    def refresh_custom(self):

        # 0 Initialization
        super().refresh_custom()
 
        if self.plot == None:
            # 1 Initial plot setup
            self.ax.grid()
            self.plot,           = self.ax.plot(self.shared_db.iis_input_ids, self.shared_db.iis_input_val[self.dim_id], color='blue', lw=1)

        elif self.shared_db.iis_new_input is not None:
            # 2 Update plot
            val = self.shared_db.iis_new_input[self.dim_id]

            if self.min_val > self.max_val:
                self.min_val        = val - 0.001
                self.max_val        = val + 0.001
                self.val_range_changed   = True

            elif val > self.max_val: 
                self.max_val        = val
                self.val_range_changed   = True

            elif val < self.min_val: 
                self.min_val = val
                self.val_range_changed   = True

            if self.shared_db.redraw:
                self.plot.set_xdata(self.shared_db.iis_input_ids)
                self.plot.set_ydata(self.shared_db.iis_input_val[self.dim_id])

        if not self.shared_db.redraw: return

        if self.shared_db.refresh_full:
            self.plot.set_xdata(self.shared_db.iis_input_ids)
            self.plot.set_ydata(self.shared_db.iis_input_val[self.dim_id])

        # 3 Update ax limits
        self.ax.set_xlim(self.shared_db.iis_min_id, self.shared_db.iis_max_id)
        if self.val_range_changed:
            self.ax.set_ylim(self.min_val, self.max_val)
            self.val_range_changed = False

        self.set_flush_events(self.shared_db.iis_flush_events)



## -------------------------------------------------------------------------------------------------
    def assign_dim(self, p_dim_id):
        self.dim_id = p_dim_id





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class IISMain2D(SciUISubplot2D):
    """
    ...
    """
    
    C_NAME                      = 'Interactive 2D Input Space'
    C_AX_RECTANGLE              = [0.09, 0.09 ,0.88 ,0.88]
    C_AX_FRAME                  = True
    C_AX_CURSOR                 = True
            
## -------------------------------------------------------------------------------------------------
    def cb_mouse_moved(self, event):
        super().cb_mouse_moved(event)
        self.shared_db.iis_update_counter = 0
        self.shared_db.iis_flush_events   = False
        self.shared_db.iis_sim_input_cb([event.xdata, event.ydata])
        self.shared_db.iis_flush_events   = True

        
## -------------------------------------------------------------------------------------------------
    def cb_mbutton_released(self, event):
        super().cb_mbutton_released(event)
        self.shared_db.iis_update_counter = 0
        self.shared_db.iis_flush_events   = False
        self.shared_db.iis_sim_input_cb([event.xdata, event.ydata])
        self.shared_db.iis_flush_events   = True


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.axes_up_to_date    = False
        self.plot               = None


## -------------------------------------------------------------------------------------------------
    def refresh_custom(self):
        super().refresh_custom()

        if self.shared_db.refresh_full or not self.axes_up_to_date:
            self.refresh_axes()
            self.axes_up_to_date = True

        if not self.shared_db.refresh_full: return

        if self.plot is None:
            self.plot,  = self.ax.plot(self.shared_db.iis_input_val[0], self.shared_db.iis_input_val[1], 'b+')

        elif self.shared_db.redraw:
            self.plot.set_xdata(self.shared_db.iis_input_val[0])
            self.plot.set_ydata(self.shared_db.iis_input_val[1])

        self.set_flush_events(self.shared_db.iis_flush_events)
        

## -------------------------------------------------------------------------------------------------
    def refresh_axes(self):
        if not self.shared_db.redraw: return
        
        ids     = self.shared_db.iis_ispace.get_dim_ids()
        dim0    = self.shared_db.iis_ispace.get_dim(ids[0]) 
        dim1    = self.shared_db.iis_ispace.get_dim(ids[1]) 
        
        self.ax.set_xlabel('Input Variable ' + dim0.get_name_short())
        self.ax.set_ylabel('Input Variable ' + dim1.get_name_short())
        self.ax.set_xbound(dim0.get_boundaries())
        self.ax.set_ybound(dim1.get_boundaries())
        self.ax.set_xlim(dim0.get_boundaries())
        self.ax.set_ylim(dim1.get_boundaries())
        self.ax.grid(True)





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class IISMain3D(SciUISubplot3D):
    """
    ...
    """
    
    C_NAME                      = 'Interactive 3D Input Space'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.axes_up_to_date    = False
        self.plot               = None


## -------------------------------------------------------------------------------------------------
    def refresh_custom(self):
        super().refresh_custom()

        if not self.axes_up_to_date:
            self.refresh_axes()
            self.axes_up_to_date = True

        if self.shared_db.iis_new_input is None: return

        if self.plot is None:
            # Create an object of type mpl_toolkits.mplot3d.art3d.Lines3D...
            self.plot, = self.ax.plot3D(self.shared_db.iis_input_val[0], self.shared_db.iis_input_val[1], self.shared_db.iis_input_val[2], color='blue', linestyle='None', marker='+')

        else:
            self.plot.set_data_3d(self.shared_db.iis_input_val[0], self.shared_db.iis_input_val[1], self.shared_db.iis_input_val[2])
 
        self.set_flush_events(self.shared_db.iis_flush_events)


## -------------------------------------------------------------------------------------------------
    def refresh_axes(self):
        if not self.shared_db.refresh_full: return 

        dim0    = self.shared_db.iis_ispace.get_dim(0) 
        dim1    = self.shared_db.iis_ispace.get_dim(1) 
        dim2    = self.shared_db.iis_ispace.get_dim(2) 
        
        self.ax.set_xlabel('Input Variable ' + dim0.get_name_short())
        self.ax.set_ylabel('Input Variable ' + dim1.get_name_short())
        self.ax.set_zlabel('Input Variable ' + dim2.get_name_short())
        self.ax.set_xbound(dim0.get_boundaries())
        self.ax.set_ybound(dim1.get_boundaries())
        self.ax.set_zbound(dim2.get_boundaries())
        self.ax.set_xlim(dim0.get_boundaries())
        self.ax.set_ylim(dim1.get_boundaries())
        self.ax.set_zlim(dim2.get_boundaries())





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class IISParamGeneral(SciUIFrameParam):
    """
    ...
    """

    C_NAME                      = 'General'
    C_WSIZE_VALUES              = [0, 10, 50, 100, 200, 500, 1000, 2000]
    C_URATE_VALUES              = [1, 2, 5, 10, 50, 100]


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()

        self.first_refresh  = True
        self.wsize_var      = IntVar()
        self.wsize_var.set(self.shared_db.iis_input_horizon)
        self.urate_var      = IntVar()
        self.urate_var.set(self.shared_db.iis_update_step)


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        if self.first_refresh: 
            self.first_refresh = False
        else: 
            return

        super().refresh(p_parent_frame=p_parent_frame)

        # 1 Window size
        wsize_label = Label(self.frame, text='Window size           ', font=self.fontstyle)     
        wsize_label.grid(row=0, column=0, sticky='w')
        self.wsize  = Combobox(self.frame, textvariable=self.wsize_var, width=6, justify='right', state='normal', background='white', font=self.fontstyle, values=self.C_WSIZE_VALUES)
        self.wsize.grid(row=0, column=1)
        self.wsize_var.trace('w', self.__cb_wsize_changed)
        Label(self.frame, text='                                                                                                  ').grid(row=0, column=2)

        # 2 Update rate (with tool tip)
        urate_label = Label(self.frame, text='Update rate', font=self.fontstyle)
        urate_label.grid(row=1, column=0, sticky='w')
        self.urate  = Combobox(self.frame, textvariable=self.urate_var, width=6, justify='right', state='normal', background='white', font=self.fontstyle, values=[1,2,5,10,20,50,100,200,500,1000])
        self.urate.grid(row=1, column=1)
        self.urate_var.trace('w', self.__cb_urate_changed)
        SciUITooltip(self.urate, 'Number of inputs after which the screen is updated')


## -------------------------------------------------------------------------------------------------
    def __cb_wsize_changed(self, *args):
        try:
            wsize   = self.wsize_var.get()
        except:
            return

        if wsize == 1: 
            wsize = 2
            self.wsize_var.set(wsize)

        self.shared_db.iis_input_horizon    = wsize
        self.shared_db.iis_update_counter   = 1
        self.log(self.C_LOG_TYPE_I, 'Data window size changed to', str(self.shared_db.iis_input_horizon))

        self.shared_db.refresh_full         = True
        self.shared_db.start_global_refresh()



 ## -------------------------------------------------------------------------------------------------
    def __cb_urate_changed(self, *args):
        try:
            ustep = self.urate_var.get()
        except:
            return

        if ustep < 1:
            ustep = 1
            self.urate_var.set(ustep)

        self.shared_db.iis_update_step      = ustep
        self.shared_db.iis_update_counter   = 1

        self.log(self.C_LOG_TYPE_I, 'Screen update rate changed to', str(self.shared_db.iis_update_step))






## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class IISParamBenchmark(SciUIFrameParam):
    """
    ...
    """

    C_NAME                      = 'Benchmark Test'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
 
        self.first_refresh      = True
        self.benchmark_var      = StringVar()
        self.benchmarks_cbox    = []
        self.benchmarks         = []
        self.num_inputs_var     = IntVar()
        self.width_cbox         = 55    # 80

        if self.shared_db.iis_ispace.get_num_dim() == 2:
            benchmark_classes = IISBenchmark2D.__subclasses__()

        elif self.shared_db.iis_ispace.get_num_dim() == 3:
            benchmark_classes = IISBenchmark3D.__subclasses__()

        for bm_class in benchmark_classes:
            bm_key = bm_class.C_GROUP + ' - ' + bm_class.C_NAME
            self.benchmarks.append([bm_key, bm_class])
            self.benchmarks_cbox.append(bm_key)

        self.benchmarks_cbox.sort()


## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        if self.first_refresh: 
            self.first_refresh = False
        else: 
            return

        super().refresh(p_parent_frame=p_parent_frame)

        # 1 Combo box for benchmark selection
        Label(self.frame, text='Benchmark Test     ', pady=3, font=self.fontstyle).grid(row=0, column=0, sticky='W')
        self.benchmark_cbox = Combobox(self.frame, textvariable=self.benchmark_var, values=self.benchmarks_cbox, state='readonly')
        self.benchmark_cbox.config(width=self.width_cbox, font=self.fontstyle)
        self.benchmark_cbox.grid(row=0, column=1, sticky='w')
        self.benchmark_var.set(self.benchmarks_cbox[0])
        self.benchmark_var.trace('w', self.__cb_benchmark_changed)

        # 2 Text box for benchmark description
        Label(self.frame, text='Description', pady=3, font=self.fontstyle).grid(row=1, column=0, sticky='NW')
        self.description    = Text(self.frame, height=5, width=self.width_cbox-5)
        self.description.grid(row=1, column=1, sticky='w')

        # 3 Info field for proposed window size/horizon

         
        # 4 Info field for number of inputs to be processed 
        Label(self.frame, text='Number of inputs', pady=3, font=self.fontstyle).grid(row=2, column=0, sticky='NW')
        self.num_inputs  = Entry(self.frame, textvariable=self.num_inputs_var, width=8, justify='right', font=self.fontstyle)
        self.num_inputs.grid(row=2, column=1, sticky='w')


        Label(self.frame, text=' ', pady=3, width=self.width_cbox+2, font=self.fontstyle).grid(row=3, column=1, sticky='W')

        # 5 Button for submitting the benchmark test
        self.submit = Button(self.frame, text='Run', command=self.__cb_benchmark_run)
        self.submit.grid(row=4, column=1, sticky='E')

        # 6 Update detail fields
        self.__benchmark_show_details(self.benchmarks_cbox[0])


## -------------------------------------------------------------------------------------------------
    def __benchmark_show_details(self, p_bm):
        for bm in self.benchmarks:
            if bm[0] == p_bm: break

        self.description.delete("1.0","end")
        self.description.insert(tkinter.END, bm[1].C_DESCRIPTION)
        self.num_inputs_var.set(bm[1].C_INPUTS)


## -------------------------------------------------------------------------------------------------
    def __cb_benchmark_changed(self, *args):
        benchmark   = self.benchmark_var.get()
        self.log(self.C_LOG_TYPE_I, 'Benchmark changed to "' + benchmark + '"')
        self.__benchmark_show_details(benchmark)


## -------------------------------------------------------------------------------------------------
    def __cb_benchmark_run(self, *args):
        benchmark   = self.benchmark_var.get()
 
        for bm in self.benchmarks:
            if bm[0] == benchmark: break

        self.log(self.C_LOG_TYPE_I, 'Started benchmark "' + benchmark + '"')
        bm[1](self.shared_db.iis_ispace, self.shared_db.iis_sim_input_cb, p_logging=self._level).run()
        self.shared_db.iis_update_counter = 0
        self.shared_db.start_global_refresh()

       



## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class IISParam(SciUIFrameParam):
    """
    Main parameter frame of IIS.
    """

    C_NAME                      = 'Settings'

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()

        self.add_component(IISParamGeneral(p_shared_db=self.shared_db, p_row=0, p_col=0, p_pady = 5, p_visible=True, p_logging=self._level))
        self.add_component(IISParamBenchmark(p_shared_db=self.shared_db, p_row=1, p_col=0, p_pady = 5, p_visible=True, p_logging=self._level))





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class IISFrameLeft(SciUIFrame):
    """
    ...
    """

    C_NAME                      = 'IIS Frame left'
    C_HEIGHT_VAR_PLOTS          = 0.5

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()

        # 2 Create subplots for each input space dimension
        num_dim     = self.shared_db.iis_ispace.get_num_dim() 
        height_perc = self.C_HEIGHT_VAR_PLOTS / num_dim

        for i, dim_id in enumerate(self.shared_db.iis_ispace.get_dim_ids()):
            dim = self.shared_db.iis_ispace.get_dim(dim_id)
            iisivar_comp = IISIVar(self.shared_db, p_row=i, p_col=0, p_title='Input Variable ' + dim.get_name_short(), p_width_perc=0.25, p_height_perc=height_perc, p_logging=self._level)
            iisivar_comp.assign_dim(i)
            self.add_component(iisivar_comp)

        self.add_component(IISParam(self.shared_db, p_row=num_dim, p_col=0, p_visible=True, p_logging=self._level))





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class InteractiveInputSpace(SciUIFrame): 
    """
    Main class for the interactive 2D/3D input space.
    """

    C_NAME                      = 'Interactive Input Space'

## -------------------------------------------------------------------------------------------------
    @staticmethod
    def enrich_shared_db(p_shared_db:SciUISharedDB):
        p_shared_db.iis_ispace          = ESpace()
        p_shared_db.iis_new_input       = None
        p_shared_db.iis_input_id        = -1
        p_shared_db.iis_input_num       = 0
        p_shared_db.iis_input_ids       = []
        p_shared_db.iis_input_val       = [[],[],[]]
        p_shared_db.iis_input_horizon   = 100
        p_shared_db.iis_min_id          = 0
        p_shared_db.iis_max_id          = p_shared_db.iis_input_horizon - 1
        p_shared_db.iis_update_step     = 1
        p_shared_db.iis_update_counter  = 1
        p_shared_db.iis_flush_events    = True


## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()
        self.shared_db.iis_sim_input_cb = self.simulate_input


        # 1 Detect input space dimensionality
        try:
            self.dim = self.shared_db.iis_ispace.get_num_dim()
            self.log(self.C_LOG_TYPE_I, 'Input space with', str(self.dim), 'dimensions detected')
            if ( self.dim != 2 ) and ( self.dim != 3 ):
                self.log(self.C_LOG_TYPE_E, 'class is suitable for 2D and 3D input spaces only!') 
                return
        except:
            self.log(self.C_LOG_TYPE_E, 'Missing input space object (self.shared_db.iis_ispace)!')
            return


        # 2 Create left subframe with subplots for each input dimension
        self.add_component(IISFrameLeft(self.shared_db, p_row=0, p_col=0, p_padx=5, p_logging=self._level))


        #  3 Create input space subplot
        if self.dim == 2:
            self.add_component(IISMain2D(self.shared_db, p_row=0, p_col=1, p_height_perc=1.0, p_padx=5, p_logging=self._level))
        elif self.dim == 3:
            self.add_component(IISMain3D(self.shared_db, p_row=0, p_col=1, p_height_perc=1.0, p_padx=5, p_logging=self._level))


 ## -------------------------------------------------------------------------------------------------
    def refresh(self, p_parent_frame):
        if self.shared_db.refresh_full: self.update_data()

        if self.shared_db.iis_update_counter <= 1:
            self.shared_db.iis_update_counter = self.shared_db.iis_update_step
            self.shared_db.redraw = True
        else:
            self.shared_db.iis_update_counter -= 1
            self.shared_db.redraw = False

        super().refresh(p_parent_frame=p_parent_frame)


 ## -------------------------------------------------------------------------------------------------
    def update_data(self):
        self.shared_db.iis_max_id = max(1, self.shared_db.iis_input_id, self.shared_db.iis_input_horizon)      

        if self.shared_db.iis_input_horizon > 0: 
            if self.shared_db.iis_max_id > self.shared_db.iis_input_horizon:
                self.shared_db.iis_min_id = self.shared_db.iis_max_id - self.shared_db.iis_input_horizon + 1
            else:
                self.shared_db.iis_min_id = 0

            num_pop = max(len(self.shared_db.iis_input_ids) - self.shared_db.iis_input_horizon, 0)

            for i in range(num_pop): 
                self.shared_db.iis_input_ids.pop(0)

                for d in range(self.shared_db.iis_ispace.get_num_dim()):
                    self.shared_db.iis_input_val[d].pop(0)


## -------------------------------------------------------------------------------------------------
    def simulate_input(self, p_x): 
        self.log(self.C_LOG_TYPE_I, 'Simulate input', p_x)
        
        # 0 Initialization
        self.shared_db.iis_new_input    = p_x
        horizon                         = self.shared_db.iis_input_horizon
        ids                             = self.shared_db.iis_input_ids
        values                          = self.shared_db.iis_input_val

        # 1 Update input ids
        self.shared_db.iis_input_id     += 1
        ids.append(self.shared_db.iis_input_id)
        self.shared_db.iis_max_id       = max(1, self.shared_db.iis_input_id, horizon)      

        # 2 Append new input components to value lists
        for d, val in enumerate(p_x): values[d].append(val)

        # 3 Start global refresh
        self.shared_db.refresh_full = True
        self.shared_db.start_global_refresh()   

        # 4 Reset shared input variable
        self.shared_db.iis_new_input    = None         
